#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import warnings
from copy import deepcopy
from math import sqrt
from typing import Any, SupportsIndex

import numpy as np
from ax.adapter.base import Adapter, DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.transforms.winsorize import (
    _get_auto_winsorization_cutoffs_outcome_constraint,
    _get_auto_winsorization_cutoffs_single_objective,
    _get_tukey_cutoffs,
    AUTO_WINS_QUANTILE,
    Winsorize,
)
from ax.core.arm import Arm
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.metric import Metric
from ax.core.objective import MultiObjective, Objective, ScalarizedObjective
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import (
    ComparisonOp,
    ObjectiveThreshold,
    OutcomeConstraint,
    ScalarizedOutcomeConstraint,
)
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.exceptions.core import AxOptimizationWarning, DataRequiredError, UserInputError
from ax.generators.base import Generator
from ax.generators.winsorization_config import WinsorizationConfig
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_experiment_with_observations,
    get_observations_with_invalid_value,
    get_optimization_config,
)
from pandas import DataFrame
from pandas.testing import assert_frame_equal

INF = float("inf")
OBSERVATION_DATA = [
    Observation(
        features=ObservationFeatures(parameters={"x": 2.0, "y": 10.0}),
        data=ObservationData(
            means=np.array([1.0, 2.0, 6.0]),
            covariance=np.array([[1.0, 2.0, 0.0], [3.0, 4.0, 0.0], [0.0, 0.0, 4.0]]),
            metric_names=["a", "b", "b"],
        ),
        arm_name="1_1",
    )
]


class WinsorizeTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.obsd1 = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([0.0, 0.0, 1.0]),
            covariance=np.array([[1.0, 0.2, 0.4], [0.2, 2.0, 0.8], [0.4, 0.8, 3.0]]),
        )
        self.obsd2 = ObservationData(
            metric_names=["m1", "m1", "m2", "m2"],
            means=np.array([1.0, 2.0, 2.0, 1.0]),
            covariance=np.array(
                [
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.2, 0.4],
                    [0.0, 0.2, 2.0, 0.8],
                    [0.0, 0.4, 0.8, 3.0],
                ]
            ),
        )
        self.observations = [
            Observation(features=ObservationFeatures({}), data=obsd)
            for obsd in [self.obsd1, self.obsd2]
        ]
        self.experiment_data = extract_experiment_data(
            experiment=get_experiment_with_observations(
                observations=[  # Same means as above.
                    [0.0, 0.0],
                    [0.0, 1.0],
                    [1.0, 2.0],
                    [2.0, 1.0],
                ],
                sems=[
                    [1.0, sqrt(2.0)],
                    [1.0, sqrt(3.0)],
                    [1.0, sqrt(2.0)],
                    [1.0, sqrt(3.0)],
                ],
            ),
            data_loader_config=DataLoaderConfig(),
        )

        self.t = Winsorize(
            search_space=None,
            experiment_data=self.experiment_data,
            config={
                "winsorization_config": WinsorizationConfig(upper_quantile_margin=0.2)
            },
        )
        self.t1 = Winsorize(
            search_space=None,
            observations=deepcopy(self.observations),
            config={
                "winsorization_config": WinsorizationConfig(upper_quantile_margin=0.8)
            },
        )
        self.t2 = Winsorize(
            search_space=None,
            observations=deepcopy(self.observations),
            config={
                "winsorization_config": WinsorizationConfig(lower_quantile_margin=0.2)
            },
        )
        self.t3 = Winsorize(
            search_space=None,
            observations=deepcopy(self.observations),
            config={
                "winsorization_config": {
                    "m1": WinsorizationConfig(upper_quantile_margin=0.6),
                    "m2": WinsorizationConfig(
                        upper_quantile_margin=0.6, upper_boundary=1.9
                    ),
                }
            },
        )
        self.t4 = Winsorize(
            search_space=None,
            observations=deepcopy(self.observations),
            config={
                "winsorization_config": {
                    "m1": WinsorizationConfig(lower_quantile_margin=0.8),
                    "m2": WinsorizationConfig(
                        lower_quantile_margin=0.8, lower_boundary=0.3
                    ),
                }
            },
        )

        self.obsd3 = ObservationData(
            metric_names=["m3", "m3", "m3", "m3"],
            means=np.array([0.0, 1.0, 5.0, 3.0]),
            covariance=np.eye(4),
        )
        self.obs3 = Observation(features=ObservationFeatures({}), data=self.obsd3)
        self.t5 = Winsorize(
            search_space=None,
            observations=deepcopy(self.observations) + deepcopy([self.obs3]),
            config={
                "winsorization_config": {
                    "m1": WinsorizationConfig(upper_quantile_margin=0.6),
                    "m2": WinsorizationConfig(lower_quantile_margin=0.4),
                }
            },
        )
        self.t6 = Winsorize(
            search_space=None,
            observations=deepcopy(self.observations),
            config={
                "winsorization_config": {
                    "m1": WinsorizationConfig(upper_quantile_margin=0.6),
                    "m2": WinsorizationConfig(
                        lower_quantile_margin=0.4, lower_boundary=0.0
                    ),
                }
            },
        )

    def test_Init(self) -> None:
        self.assertEqual(self.t.cutoffs["m1"], (-INF, 2.0))
        self.assertEqual(self.t.cutoffs["m2"], (-INF, 2.0))
        self.assertEqual(self.t1.cutoffs["m1"], (-INF, 1.0))
        self.assertEqual(self.t1.cutoffs["m2"], (-INF, 1.0))
        self.assertEqual(self.t2.cutoffs["m1"], (0.0, INF))
        self.assertEqual(self.t2.cutoffs["m2"], (0.0, INF))
        with self.assertRaisesRegex(
            DataRequiredError,
            "`Winsorize` transform requires non-empty data.",
        ):
            Winsorize(search_space=None, observations=[])
        with self.assertRaisesRegex(
            UserInputError,
            "Transform config for `Winsorize` transform must be specified and "
            "non-empty when using winsorization.",
        ):
            Winsorize(
                search_space=None,
                observations=deepcopy(self.observations[:1]),
            )
        with self.assertRaisesRegex(
            UserInputError,
            "`derelativize_with_raw_status_quo` must be a boolean. Got 1234.",
        ):
            Winsorize(
                search_space=None,
                observations=deepcopy(self.observations[:1]),
                config={"derelativize_with_raw_status_quo": 1234},
            )

    def test_TransformObservations(self) -> None:
        observation_data = self.t1._transform_observation_data([deepcopy(self.obsd1)])[
            0
        ]
        self.assertListEqual(list(observation_data.means), [0.0, 0.0, 1.0])
        observation_data = self.t1._transform_observation_data([deepcopy(self.obsd2)])[
            0
        ]
        self.assertListEqual(list(observation_data.means), [1.0, 1.0, 1.0, 1.0])
        observation_data = self.t2._transform_observation_data([deepcopy(self.obsd1)])[
            0
        ]
        self.assertListEqual(list(observation_data.means), [0.0, 0.0, 1.0])
        observation_data = self.t2._transform_observation_data([deepcopy(self.obsd2)])[
            0
        ]
        self.assertListEqual(list(observation_data.means), [1.0, 2.0, 2.0, 1.0])

    def test_InitPercentileBounds(self) -> None:
        self.assertEqual(self.t3.cutoffs["m1"], (-INF, 1.0))
        self.assertEqual(self.t3.cutoffs["m2"], (-INF, 1.9))
        self.assertEqual(self.t4.cutoffs["m1"], (1.0, INF))
        self.assertEqual(self.t4.cutoffs["m2"], (0.3, INF))

    def test_TransformObservationsPercentileBounds(self) -> None:
        observation_data = self.t3._transform_observation_data([deepcopy(self.obsd1)])[
            0
        ]
        self.assertListEqual(list(observation_data.means), [0.0, 0.0, 1.0])
        observation_data = self.t3._transform_observation_data([deepcopy(self.obsd2)])[
            0
        ]
        self.assertListEqual(list(observation_data.means), [1.0, 1.0, 1.9, 1.0])
        observation_data = self.t4._transform_observation_data([deepcopy(self.obsd1)])[
            0
        ]
        self.assertListEqual(list(observation_data.means), [1.0, 0.3, 1.0])
        observation_data = self.t4._transform_observation_data([deepcopy(self.obsd2)])[
            0
        ]
        self.assertListEqual(list(observation_data.means), [1.0, 2.0, 2.0, 1.0])

    def test_TransformObservationsDifferentLowerUpper(self) -> None:
        observation_data = self.t5._transform_observation_data([deepcopy(self.obsd2)])[
            0
        ]
        self.assertEqual(self.t5.cutoffs["m1"], (-INF, 1.0))
        self.assertEqual(self.t5.cutoffs["m2"], (1.0, INF))
        self.assertEqual(self.t5.cutoffs["m3"], (-INF, INF))
        self.assertListEqual(list(observation_data.means), [1.0, 1.0, 2.0, 1.0])
        # Nothing should happen to m3
        observation_data = self.t5._transform_observation_data([deepcopy(self.obsd3)])[
            0
        ]
        self.assertListEqual(list(observation_data.means), [0.0, 1.0, 5.0, 3.0])
        # With winsorization boundaries
        observation_data = self.t6._transform_observation_data([deepcopy(self.obsd2)])[
            0
        ]
        self.assertEqual(self.t6.cutoffs["m1"], (-INF, 1.0))
        self.assertEqual(self.t6.cutoffs["m2"], (0.0, INF))
        self.assertListEqual(list(observation_data.means), [1.0, 1.0, 2.0, 1.0])

    def test_optimization_config_default(self) -> None:
        # Specify the winsorization
        optimization_config = get_optimization_config()
        percentiles = get_default_transform_cutoffs(
            optimization_config=optimization_config,
            winsorization_config={"m1": WinsorizationConfig(0.2, 0.0)},
        )
        self.assertDictEqual(percentiles, {"m1": (1, INF)})

    def test_tukey_cutoffs(self) -> None:
        Y = np.array([-100, 0, 1, 2, 50])
        self.assertEqual(_get_tukey_cutoffs(Y=Y, lower=True), -3.0)
        self.assertEqual(_get_tukey_cutoffs(Y=Y, lower=False), 5.0)

    def test_winsorize_outcome_constraints(self) -> None:
        metric_values = [-100, 0, 1, 2, 3, 4, 5, 6, 7, 50]
        ma, mb = Metric(name="a"), Metric(name="b")
        outcome_constraint_leq = OutcomeConstraint(
            metric=ma, op=ComparisonOp.LEQ, bound=10, relative=False
        )
        outcome_constraint_geq = OutcomeConstraint(
            metric=mb, op=ComparisonOp.GEQ, bound=-9, relative=False
        )
        # From above with a loose bound
        cutoffs = _get_auto_winsorization_cutoffs_outcome_constraint(
            # pyre-fixme[6]: For 1st param expected `List[float]` but got `List[int]`.
            metric_values=metric_values,
            outcome_constraints=[outcome_constraint_leq],
        )
        self.assertEqual(cutoffs, (-INF, 23.5))
        # From above with a tight bound
        outcome_constraint_leq.bound = 2
        cutoffs = _get_auto_winsorization_cutoffs_outcome_constraint(
            # pyre-fixme[6]: For 1st param expected `List[float]` but got `List[int]`.
            metric_values=metric_values,
            outcome_constraints=[outcome_constraint_leq],
        )
        self.assertEqual(cutoffs, (-INF, 13.5))
        # From below with a loose bound
        cutoffs = _get_auto_winsorization_cutoffs_outcome_constraint(
            # pyre-fixme[6]: For 1st param expected `List[float]` but got `List[int]`.
            metric_values=metric_values,
            outcome_constraints=[outcome_constraint_geq],
        )
        self.assertEqual(cutoffs, (-31.5, INF))
        # From below with a tight bound
        outcome_constraint_geq.bound = 5
        cutoffs = _get_auto_winsorization_cutoffs_outcome_constraint(
            # pyre-fixme[6]: For 1st param expected `List[float]` but got `List[int]`.
            metric_values=metric_values,
            outcome_constraints=[outcome_constraint_geq],
        )
        self.assertEqual(cutoffs, (-6.5, INF))
        # Both with the tight bounds
        outcome_constraint_geq.bound = 5
        cutoffs = _get_auto_winsorization_cutoffs_outcome_constraint(
            # pyre-fixme[6]: For 1st param expected `List[float]` but got `List[int]`.
            metric_values=metric_values,
            outcome_constraints=[outcome_constraint_leq, outcome_constraint_geq],
        )
        self.assertEqual(cutoffs, (-6.5, 13.5))

    def test_winsorization_single_objective(self) -> None:
        metric_values = [-100.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 50.0]
        cutoffs = _get_auto_winsorization_cutoffs_single_objective(
            metric_values=metric_values, minimize=True
        )
        self.assertEqual(cutoffs, (-INF, 13.5))
        cutoffs = _get_auto_winsorization_cutoffs_single_objective(
            metric_values=metric_values, minimize=False
        )
        self.assertEqual(cutoffs, (-6.5, INF))

    def test_winsorization_without_optimization_config(self) -> None:
        means = np.array([-100, 0, 1, 2, 3, 4, 5, 6, 7, 50])
        obsd = ObservationData(
            metric_names=["m1"] * 10,
            means=means,
            covariance=np.eye(10),
        )
        config = {
            "winsorization_config": {
                # pyre-fixme[6]: For 1st param expected `float` but got `None`.
                # pyre-fixme[6]: For 2nd param expected `float` but got `None`.
                "m1": WinsorizationConfig(None, None),
            }
        }
        transform = get_transform(observation_data=[deepcopy(obsd)], config=config)
        self.assertEqual(transform.cutoffs["m1"], (-INF, INF))
        # None and 0.0 should be treated the same way
        config["winsorization_config"]["m1"] = WinsorizationConfig(0.0, 0.0)
        transform = get_transform(observation_data=[deepcopy(obsd)], config=config)
        self.assertEqual(transform.cutoffs["m1"], (-INF, INF))
        # From above
        config["winsorization_config"]["m1"] = WinsorizationConfig(0.0, 0.2)
        transform = get_transform(observation_data=[deepcopy(obsd)], config=config)
        self.assertEqual(transform.cutoffs["m1"], (-INF, 7))
        # From below
        config["winsorization_config"]["m1"] = WinsorizationConfig(0.2, 0.0)
        transform = get_transform(observation_data=[deepcopy(obsd)], config=config)
        self.assertEqual(transform.cutoffs["m1"], (0, INF))
        # Do both automatically
        config["winsorization_config"]["m1"] = WinsorizationConfig(
            AUTO_WINS_QUANTILE, AUTO_WINS_QUANTILE
        )
        transform = get_transform(observation_data=[deepcopy(obsd)], config=config)
        self.assertEqual(transform.cutoffs["m1"], (-6.5, 13.5))
        # Add a second metric that shouldn't be winsorized
        config["winsorization_config"]["m1"] = WinsorizationConfig(
            0.0, AUTO_WINS_QUANTILE
        )
        obsd2 = ObservationData(
            metric_names=["m2"] * 10,
            means=means,
            covariance=np.eye(10),
        )
        transform = get_transform(
            observation_data=[deepcopy(obsd), deepcopy(obsd2)], config=config
        )
        self.assertEqual(transform.cutoffs["m1"], (-INF, 13.5))
        self.assertEqual(transform.cutoffs["m2"], (-INF, INF))
        # Winsorize both
        config["winsorization_config"]["m2"] = WinsorizationConfig(0.2, 0.0)
        transform = get_transform(
            observation_data=[deepcopy(obsd), deepcopy(obsd2)], config=config
        )
        self.assertEqual(transform.cutoffs["m1"], (-INF, 13.5))
        self.assertEqual(transform.cutoffs["m2"], (0.0, INF))

    def test_winsorization_with_optimization_config(self) -> None:
        obsd_1 = ObservationData(
            metric_names=["m1"] * 10,
            means=np.array([-100, 0, 1, 2, 3, 4, 5, 6, 7, 50]),
            covariance=np.eye(10),
        )
        obsd_2 = ObservationData(
            metric_names=["m2"] * 7,
            means=np.array([-10, 0, 1, 2, 3, 4, 47]),
            covariance=np.eye(7),
        )
        obsd_3 = ObservationData(
            metric_names=["m3"] * 6,
            means=np.array([-456, 1, 2, 3, 4, 9]),
            covariance=np.eye(6),
        )
        all_obsd = [obsd_1, obsd_2, obsd_3]
        m1 = Metric(name="m1", lower_is_better=False)
        m2 = Metric(name="m2", lower_is_better=True)
        m3 = Metric(name="m3")
        # Scalarized objective
        for minimize in [True, False]:
            optimization_config = OptimizationConfig(
                objective=ScalarizedObjective(
                    metrics=[Metric(name="m1"), Metric(name="m2")],
                    weights=[1, -1],
                    minimize=minimize,
                )
            )
            transform = get_transform(
                observation_data=deepcopy(all_obsd),
                optimization_config=optimization_config,
            )
            if minimize:
                self.assertEqual(
                    transform.cutoffs,
                    {"m1": (-INF, 13.5), "m2": (-6.0, INF), "m3": (-INF, INF)},
                )
            else:
                self.assertEqual(
                    transform.cutoffs,
                    {"m1": (-6.5, INF), "m2": (-INF, 10.0), "m3": (-INF, INF)},
                )
        # Simple single-objective problem
        optimization_config = OptimizationConfig(
            objective=Objective(metric=m2, minimize=True)
        )
        transform = get_transform(
            observation_data=deepcopy(all_obsd), optimization_config=optimization_config
        )
        self.assertEqual(transform.cutoffs["m1"], (-INF, INF))
        self.assertEqual(transform.cutoffs["m2"], (-INF, 10.0))  # 4 + 1.5 * 4
        self.assertEqual(transform.cutoffs["m3"], (-INF, INF))
        # Add a relative constraint, which should warn and skip relative metrics.
        outcome_constraint = OutcomeConstraint(
            metric=m1, op=ComparisonOp.LEQ, bound=3, relative=True
        )
        optimization_config = OptimizationConfig(
            objective=Objective(metric=m2, minimize=True),
            outcome_constraints=[outcome_constraint],
        )
        with self.assertWarnsRegex(
            AxOptimizationWarning,
            "Automatic winsorization doesn't support relative outcome constraints "
            "or objective thresholds when `derelativize_with_raw_status_quo` is not "
            "set to `True`.",
        ):
            transform = get_transform(
                observation_data=deepcopy(all_obsd),
                optimization_config=optimization_config,
            )
        self.assertEqual(transform.cutoffs["m1"], (-INF, INF))
        self.assertEqual(transform.cutoffs["m2"], (-INF, 10.0))  # 4 + 1.5 * 4
        self.assertEqual(transform.cutoffs["m3"], (-INF, INF))
        # Make the constraint absolute, which should trigger winsorization
        optimization_config.outcome_constraints[0].relative = False
        transform = get_transform(
            observation_data=deepcopy(all_obsd), optimization_config=optimization_config
        )
        self.assertEqual(transform.cutoffs["m1"], (-INF, 13.5))  # 6 + 1.5 * 5
        self.assertEqual(transform.cutoffs["m2"], (-INF, 10.0))  # 4 + 1.5 * 4
        self.assertEqual(transform.cutoffs["m3"], (-INF, INF))
        # Change to a GEQ constraint
        optimization_config.outcome_constraints[0].op = ComparisonOp.GEQ
        transform = get_transform(
            observation_data=deepcopy(all_obsd), optimization_config=optimization_config
        )
        self.assertEqual(transform.cutoffs["m1"], (-6.5, INF))  # 1 - 1.5 * 5
        self.assertEqual(transform.cutoffs["m2"], (-INF, 10.0))  # 4 + 1.5 * 4
        self.assertEqual(transform.cutoffs["m3"], (-INF, INF))
        # Add a scalarized outcome constraint which should print a warning
        optimization_config.outcome_constraints = [
            ScalarizedOutcomeConstraint(
                metrics=[m1, m3], op=ComparisonOp.GEQ, bound=3, relative=False
            )
        ]
        with warnings.catch_warnings(record=True) as ws:
            transform = get_transform(
                observation_data=deepcopy(all_obsd),
                optimization_config=optimization_config,
            )
        for i in range(2):
            self.assertTrue(
                "Automatic winsorization isn't supported for a "
                "`ScalarizedOutcomeConstraint`. Specify the winsorization settings "
                f"manually if you want to winsorize metric m{['1', '3'][i]}."
                in [str(w.message) for w in ws]
            )
        # Multi-objective without objective thresholds should warn and winsorize
        moo_objective = MultiObjective(
            [Objective(m1, minimize=False), Objective(m2, minimize=True)]
        )
        optimization_config = MultiObjectiveOptimizationConfig(objective=moo_objective)
        with warnings.catch_warnings(record=True) as ws:
            transform = get_transform(
                observation_data=deepcopy(all_obsd),
                optimization_config=optimization_config,
            )
        for _ in range(2):
            self.assertTrue(
                "Encountered a `MultiObjective` without objective thresholds. We "
                "will winsorize each objective separately. We strongly recommend "
                "specifying the objective thresholds when using multi-objective "
                "optimization." in [str(w.message) for w in ws]
            )
        self.assertEqual(transform.cutoffs["m1"], (-6.5, INF))
        self.assertEqual(transform.cutoffs["m2"], (-INF, 10.0))
        self.assertEqual(transform.cutoffs["m3"], (-INF, INF))
        # Add relative objective thresholds. Should warn and skip.
        objective_thresholds = [
            ObjectiveThreshold(m1, 3, relative=True),
            ObjectiveThreshold(m2, 4, relative=True),
        ]
        optimization_config = MultiObjectiveOptimizationConfig(
            objective=moo_objective,
            objective_thresholds=objective_thresholds,
            outcome_constraints=[],
        )
        with self.assertWarnsRegex(
            AxOptimizationWarning,
            "Automatic winsorization doesn't support relative outcome constraints or "
            "objective thresholds when `derelativize_with_raw_status_quo` is not set "
            "to `True`.",
        ):
            transform = get_transform(
                observation_data=deepcopy(all_obsd),
                optimization_config=optimization_config,
            )
        for i in range(1, 4):
            self.assertEqual(transform.cutoffs[f"m{i}"], (-INF, INF))
        # Make the objective thresholds absolute (should trigger winsorization)
        optimization_config.objective_thresholds[0].relative = False
        optimization_config.objective_thresholds[1].relative = False
        transform = get_transform(
            observation_data=deepcopy(all_obsd), optimization_config=optimization_config
        )
        self.assertEqual(transform.cutoffs["m1"], (-6.5, INF))  # 1 - 1.5 * 5
        self.assertEqual(transform.cutoffs["m2"], (-INF, 10.0))  # 4 + 1.5 * 4
        self.assertEqual(transform.cutoffs["m3"], (-INF, INF))
        # Add an absolute outcome constraint
        optimization_config.outcome_constraints = [
            OutcomeConstraint(metric=m3, op=ComparisonOp.GEQ, bound=3, relative=False)
        ]
        transform = get_transform(
            observation_data=deepcopy(all_obsd), optimization_config=optimization_config
        )
        self.assertEqual(transform.cutoffs["m1"], (-6.5, INF))  # 1 - 1.5 * 5
        self.assertEqual(transform.cutoffs["m2"], (-INF, 10.0))  # 4 + 1.5 * 4
        self.assertEqual(transform.cutoffs["m3"], (-3.5, INF))  # 1 - 1.5 * 3

    def test_relative_constraints(self) -> None:
        # Adapter with in-design status quo
        search_space = SearchSpace(
            parameters=[
                RangeParameter("x", ParameterType.FLOAT, 0, 20),
                RangeParameter("y", ParameterType.FLOAT, 0, 20),
            ]
        )
        objective = Objective(Metric("c"), minimize=False)

        # Test with relative constraint, in-design status quo
        oc = OptimizationConfig(
            objective=objective,
            outcome_constraints=[
                OutcomeConstraint(
                    Metric("a"), ComparisonOp.LEQ, bound=2, relative=False
                ),
                OutcomeConstraint(
                    Metric("b"), ComparisonOp.LEQ, bound=-10, relative=True
                ),
                ScalarizedOutcomeConstraint(
                    metrics=[Metric("a"), Metric("b")],
                    weights=[0.0, 1.0],
                    op=ComparisonOp.LEQ,
                    bound=-10,
                    relative=True,
                ),
            ],
        )
        experiment = Experiment(search_space=search_space, optimization_config=oc)
        adapter = Adapter(experiment=experiment, generator=Generator())
        with self.assertRaisesRegex(
            DataRequiredError, "model was not fit with status quo"
        ):
            Winsorize(
                search_space=search_space,
                observations=OBSERVATION_DATA,
                adapter=adapter,
                config={"derelativize_with_raw_status_quo": True},
            )

        sq_arm = Arm(parameters={"x": 2.0, "y": 10.0}, name="1_1")
        experiment.status_quo = sq_arm
        t = (
            experiment.new_trial()
            .add_arm(sq_arm)
            .mark_running(no_runner_required=True)
            .mark_completed()
        )
        data = Data(
            df=DataFrame.from_records(
                [
                    {
                        "arm_name": sq_arm.name,
                        "metric_name": metric_name,
                        "mean": mean,
                        "sem": sem,
                        "trial_index": t.index,
                    }
                    for metric_name, mean, sem in (("a", 1.0, 2.0), ("b", 2.0, 4.0))
                ]
            )
        )
        experiment.attach_data(data)

        adapter = Adapter(experiment=experiment, generator=Generator())
        # Warns and skips without `derelativize_with_raw_status_quo`.
        with self.assertWarnsRegex(
            AxOptimizationWarning,
            "`derelativize_with_raw_status_quo` is not set to `True`",
        ):
            t = Winsorize(
                search_space=search_space,
                observations=OBSERVATION_DATA,
                adapter=adapter,
            )
        self.assertDictEqual(t.cutoffs, {"a": (-INF, INF), "b": (-INF, INF)})
        # Winsorizes with `derelativize_with_raw_status_quo`.
        t = Winsorize(
            search_space=search_space,
            observations=OBSERVATION_DATA,
            adapter=adapter,
            config={"derelativize_with_raw_status_quo": True},
        )
        self.assertDictEqual(t.cutoffs, {"a": (-INF, 3.5), "b": (-INF, 12.0)})

    def test_non_finite_data_raises(self) -> None:
        for invalid_value in [float("nan"), float("inf")]:
            observations = get_observations_with_invalid_value(
                invalid_value=invalid_value
            )
            config: dict[str, Any] = {
                "winsorization_config": WinsorizationConfig(upper_quantile_margin=0.2)
            }
            with self.assertRaisesRegex(
                ValueError, f"Non-finite data found for metric m1: {invalid_value}"
            ):
                Winsorize(search_space=None, observations=observations, config=config)

    def test_transform_experiment_data(self) -> None:
        transformed_data = self.t.transform_experiment_data(
            experiment_data=deepcopy(self.experiment_data)
        )
        # Data is within winsorization bounds. No change.
        self.assertEqual(self.experiment_data, transformed_data)

        # Modify the cutoffs to check for winsorization.
        self.t.cutoffs["m1"] = (0.5, 1.0)
        self.t.cutoffs["m2"] = (-INF, 1.5)
        transformed_data = self.t.transform_experiment_data(
            experiment_data=deepcopy(self.experiment_data)
        )
        # Check that the data is winsorized correctly.
        expected_mean = DataFrame(
            index=transformed_data.observation_data.index,
            columns=transformed_data.observation_data["mean"].columns,
            data=[
                [0.5, 0.0],
                [0.5, 1.0],
                [1.0, 1.5],
                [1.0, 1.0],
            ],
        )
        assert_frame_equal(transformed_data.observation_data["mean"], expected_mean)
        # Sem and arm data are not modified.
        assert_frame_equal(
            transformed_data.observation_data["sem"],
            self.experiment_data.observation_data["sem"],
        )
        assert_frame_equal(transformed_data.arm_data, self.experiment_data.arm_data)


def get_transform(
    observation_data: list[ObservationData],
    config: dict[str, Any] | None = None,
    optimization_config: OptimizationConfig | None = None,
) -> Winsorize:
    observations = [
        Observation(features=ObservationFeatures({}), data=obsd)
        for obsd in observation_data
    ]
    if optimization_config is not None:
        adapter = _wrap_optimization_config_in_adapter(
            optimization_config=optimization_config
        )
        return Winsorize(
            search_space=None,
            observations=observations,
            config=config,
            adapter=adapter,
        )
    return Winsorize(
        search_space=None,
        observations=observations,
        config=config,
    )


def get_default_transform_cutoffs(
    optimization_config: OptimizationConfig,
    winsorization_config: dict[str, WinsorizationConfig] | None = None,
    obs_data_len: SupportsIndex = 6,
) -> dict[str, tuple[float, float]]:
    obsd = ObservationData(
        metric_names=["m1"] * obs_data_len,
        means=np.array(range(obs_data_len)),
        # pyre-fixme[6]: For 1st argument expected `int` but got `SupportsIndex`.
        covariance=np.eye(obs_data_len),
    )
    obs = Observation(features=ObservationFeatures({}), data=obsd)
    adapter = _wrap_optimization_config_in_adapter(
        optimization_config=optimization_config
    )
    transform = Winsorize(
        search_space=None,
        observations=[deepcopy(obs)],
        adapter=adapter,
        config={
            "winsorization_config": winsorization_config,
        },
    )
    return transform.cutoffs


def _wrap_optimization_config_in_adapter(
    optimization_config: OptimizationConfig,
) -> Adapter:
    return Adapter(
        experiment=Experiment(search_space=SearchSpace(parameters=[])),
        generator=Generator(),
        optimization_config=optimization_config,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pandas as pd
from ax.analysis.metric_summary import MetricSummary
from ax.api.client import Client
from ax.core.metric import Metric
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_offline_experiments, get_online_experiments


class TestMetricSummary(TestCase):
    def test_compute(self) -> None:
        client = Client()
        client.configure_experiment(
            name="test_experiment",
            parameters=[],
        )
        client.configure_optimization(
            objective="foo, bar", outcome_constraints=["baz <= 0.0", "foo >= 1.0"]
        )
        # TODO: Debug error raised by
        # client.configure_metrics(metrics=[IMetric(name="qux")])

        client._experiment._tracking_metrics = {"qux": Metric(name="qux")}

        analysis = MetricSummary()

        with self.assertRaisesRegex(UserInputError, "requires an `Experiment`"):
            analysis.compute()

        experiment = client._experiment
        card = analysis.compute(experiment=experiment)

        # Test metadata
        self.assertEqual(card.name, "MetricSummary")
        self.assertEqual(card.title, "MetricSummary for `test_experiment`")
        self.assertEqual(
            card.subtitle,
            "High-level summary of the `Metric`-s in this `Experiment`",
        )
        self.assertIsNotNone(card.blob)

        # Test dataframe for accuracy
        self.assertEqual(
            {*card.df.columns},
            {
                "Goal",
                "Name",
                "Type",
                "Lower is Better",
                "Bound",
            },
        )
        expected = pd.DataFrame(
            {
                "Name": ["bar", "foo", "baz", "qux"],
                "Type": ["MapMetric", "MapMetric", "MapMetric", "Metric"],
                "Goal": pd.Series(
                    ["maximize", "maximize", "constrain", "track"],
                    dtype=pd.CategoricalDtype(
                        categories=[
                            "minimize",
                            "maximize",
                            "constrain",
                            "track",
                            "None",
                        ],
                        ordered=True,
                    ),
                ),
                "Bound": ["None", ">= 1.0", "<= 0.0", "None"],
                "Lower is Better": [False, False, "None", "None"],
            }
        )
        pd.testing.assert_frame_equal(card.df, expected)

    def test_online(self) -> None:
        # Test MetricSummary can be computed for a variety of experiments which
        # resemble those we see in an online setting.

        analysis = MetricSummary()
        for experiment in get_online_experiments():
            _ = analysis.compute(experiment=experiment)

    def test_offline(self) -> None:
        # Test MetricSummary can be computed for a variety of experiments which
        # resemble those we see in an offline setting.

        analysis = MetricSummary()
        for experiment in get_offline_experiments():
            _ = analysis.compute(experiment=experiment)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import numpy as np
import pandas as pd

from ax.analysis.healthcheck.search_space_analysis import (
    boundary_proportions_message,
    search_space_boundary_proportions,
    SearchSpaceAnalysis,
    SUBTITLE_BASE,
)
from ax.core.arm import Arm
from ax.core.generator_run import GeneratorRun
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.parameter_constraint import ParameterConstraint
from ax.core.search_space import SearchSpace
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment


class TestSearchSpaceAnalysis(TestCase):
    def test_search_space_analysis(self) -> None:
        experiment = get_branin_experiment(with_trial=False)
        arms = [
            Arm(name="1_0", parameters={"x1": 0.0, "x2": 1.0}),
            Arm(name="1_1", parameters={"x1": -5.0, "x2": 0.0}),
            Arm(name="1_2", parameters={"x1": -5.0, "x2": 1.0}),
        ]
        experiment.new_batch_trial(generator_run=GeneratorRun(arms=arms))
        ssa = SearchSpaceAnalysis(trial_index=0)
        card = ssa.compute(experiment=experiment)

        self.assertEqual(card.name, "SearchSpaceAnalysis")
        self.assertEqual(card.title, "Ax search-space boundary check [Warning]")
        subtitle = SUBTITLE_BASE + (
            "\n - **Relax lower bound of `'x1'`:** Ax is frequently suggesting values "
            "at the lower bound of `'x1'`, `x1 = -5.0`. This may indicate that the "
            "optimal value of this parameter is outside this bound, in which case "
            "decreasing this lower bound would improve optimization performance. "
            "Details: 66.67% of suggested arms are on the parameter's lower bound "
            "(threshold for this alert is 50.00%)."
        )
        self.assertEqual(card.subtitle, subtitle)

        arms = [
            Arm(name="2_0", parameters={"x1": 5.0, "x2": 1.0}),
            Arm(name="2_1", parameters={"x1": 5.0, "x2": 0.0}),
            Arm(name="2_2", parameters={"x1": -5.0, "x2": 2.0}),
        ]
        experiment.new_batch_trial(generator_run=GeneratorRun(arms=arms))
        ssa = SearchSpaceAnalysis(trial_index=1)
        card = ssa.compute(experiment=experiment)
        self.assertEqual(card.name, "SearchSpaceAnalysis")
        self.assertEqual(card.title, "Ax search-space boundary check [Success]")
        self.assertEqual(
            card.subtitle, SUBTITLE_BASE + "Search space does not need to be updated."
        )

        arms = [
            Arm(name="2_0", parameters={"x1": 5.0, "x2": 1.0}),
            Arm(name="3_1", parameters={"x1": -5.0, "x2": 0.5}),
            Arm(name="2_2", parameters={"x1": -5.0, "x2": 2.0}),
        ]
        experiment.new_batch_trial(generator_run=GeneratorRun(arms=arms))
        ssa = SearchSpaceAnalysis(trial_index=2)
        card = ssa.compute(experiment=experiment)
        self.assertEqual(card.name, "SearchSpaceAnalysis")
        self.assertEqual(card.title, "Ax search-space boundary check [Warning]")
        self.assertTrue(card.subtitle is not None and "x1" in card.subtitle)

    def test_search_space_boundary_proportions(self) -> None:
        ss = SearchSpace(
            parameters=[
                RangeParameter(
                    name="float_range_1",
                    parameter_type=ParameterType.FLOAT,
                    lower=1.0,
                    upper=6.0,
                ),
                RangeParameter(
                    name="float_range_2",
                    parameter_type=ParameterType.FLOAT,
                    lower=1.0,
                    upper=3.0,
                ),
                ChoiceParameter(
                    name="choice_ordered",
                    parameter_type=ParameterType.INT,
                    values=[1, 2, 3],
                    is_ordered=True,
                ),
            ],
            parameter_constraints=[
                ParameterConstraint(
                    constraint_dict={"float_range_1": 1.0, "float_range_2": 1.0},
                    bound=4.0,
                )
            ],
        )

        parameterizations: list[dict[str, None | bool | float | int | str]] = [
            {
                "float_range_1": 1.0,
                "float_range_2": 1.0,
                "choice_ordered": 1,
            },
            {
                "float_range_1": 2.0,
                "float_range_2": 2.0,
                "choice_ordered": 1,
            },
            {"float_range_1": 1.0, "float_range_2": 3.0, "choice_ordered": 3},
        ]

        df = search_space_boundary_proportions(
            search_space=ss, parameterizations=parameterizations
        )

        dff = pd.DataFrame(
            {
                "parameter_or_constraint": [
                    ss.parameters["float_range_1"],
                    ss.parameters["float_range_1"],
                    ss.parameters["float_range_2"],
                    ss.parameters["float_range_2"],
                    ss.parameters["choice_ordered"],
                    ss.parameters["choice_ordered"],
                    ss.parameter_constraints[0],
                ],
                "boundary": [
                    "float_range_1 = 1.0",
                    "float_range_1 = 6.0",
                    "float_range_2 = 1.0",
                    "float_range_2 = 3.0",
                    "choice_ordered = 1",
                    "choice_ordered = 3",
                    "1.0*float_range_1 + 1.0*float_range_2 = 4.0",
                ],
                "proportion": [
                    2.0 / 3,
                    0.0,
                    1.0 / 3,
                    1.0 / 3,
                    2.0 / 3,
                    1.0 / 3,
                    2.0 / 3,
                ],
                "bound": [
                    "lower",
                    "upper",
                    "lower",
                    "upper",
                    "lower",
                    "upper",
                    "upper",
                ],
            }
        )
        self.assertTrue(dff.equals(df))

        # test with no parameterizations
        df = search_space_boundary_proportions(search_space=ss, parameterizations=[])
        self.assertTrue(np.all(np.equal(df.proportion.values, np.zeros(7))))
        # test with no parameterizations in SS
        df = search_space_boundary_proportions(
            search_space=ss,
            parameterizations=[
                {
                    "float_range_1": 0.0,
                    "float_range_2": 1.0,
                    "choice_ordered": 1,
                },
            ],
        )
        self.assertTrue(np.all(np.equal(df.proportion.values, np.zeros(7))))

    def test_boundary_proportions_message(self) -> None:
        float_range_1 = RangeParameter(
            name="float_range_1",
            parameter_type=ParameterType.FLOAT,
            lower=1.0,
            upper=6.0,
        )
        float_range_2 = RangeParameter(
            name="float_range_2",
            parameter_type=ParameterType.FLOAT,
            lower=1.0,
            upper=3.0,
        )
        pc = ParameterConstraint(
            constraint_dict={"float_range_1": 1.0, "float_range_2": 1.0}, bound=4.0
        )
        df = pd.DataFrame(
            {
                "parameter_or_constraint": [
                    float_range_1,
                    float_range_2,
                    pc,
                ],
                "boundary": [
                    "float_range_1 = 1.0",
                    "float_range_2 = 5.0",
                    "1.0*float_range_1 + 1.0*float_range_2 = 4.0",
                ],
                "proportion": [2.0 / 3, 0.0, 2.0 / 3],
                "bound": ["lower", "upper", "upper"],
            }
        )

        msg = boundary_proportions_message(boundary_proportions_df=df)
        self.assertTrue(
            "float_range_1" in msg
            and "ParameterConstraint(1.0*float_range_1 + 1.0*float_range_2 <= 4.0)"
            in msg
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections.abc import Iterable
from logging import Logger
from typing import Any, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.core.observation import ObservationFeatures
from ax.core.parameter import Parameter
from ax.core.search_space import SearchSpace
from ax.utils.common.logger import get_logger

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


logger: Logger = get_logger(__name__)


class MetadataToParameterMixin:
    """
    This Mixin has utilities for converting metadata from observation features
    into a parameter.

    Transform is done in-place.
    """

    _parameter_list: list[Parameter] = []
    parameters: dict[str, dict[str, Any]] = {}

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        for parameter in self._parameter_list:
            search_space.add_parameter(parameter.clone())
        return search_space

    def transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            self._transform_observation_feature(obsf)
        return observation_features

    def untransform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            obsf.metadata = obsf.metadata or {}
            _transfer(
                src=obsf.parameters,
                dst=obsf.metadata,
                keys=[p.name for p in self._parameter_list],
                target_values=[p.target_value for p in self._parameter_list],
            )
        return observation_features

    def _transform_observation_feature(self, obsf: ObservationFeatures) -> None:
        _transfer(
            src=obsf.metadata or {},
            dst=obsf.parameters,
            keys=[p.name for p in self._parameter_list],
            target_values=[p.target_value for p in self._parameter_list],
        )

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        arm_data = experiment_data.arm_data
        for name in self.parameters:
            arm_data[name] = [md.get(name) for md in arm_data["metadata"]]
        return ExperimentData(
            arm_data=arm_data,
            observation_data=experiment_data.observation_data,
        )


def _transfer(
    src: dict[str, Any],
    dst: dict[str, Any],
    keys: Iterable[str],
    target_values: list[Any],
) -> None:
    """Transfer items in-place from one dictionary to another.

    If a key is not present in the source dictionary, the target value is used.
    """
    for key, target_value in zip(keys, target_values):
        dst[key] = src.pop(key, target_value)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


class SchedulerOptions:
    raise DeprecationWarning(
        "SchedulerOptions is deprecated following renaming of the Scheduler to "
        "Orchestrator. Please use OrchestratorOptions instead; import with: "
        "`from ax.service.orchestrator_options import OrchestratorOptions`"
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import pandas as pd
from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis
from ax.analysis.healthcheck.healthcheck_analysis import (
    create_healthcheck_analysis_card,
    HealthcheckAnalysisCard,
    HealthcheckStatus,
)
from ax.core.experiment import Experiment
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import override


class ShouldGenerateCandidates(Analysis):
    def __init__(
        self,
        should_generate: bool,
        reason: str,
        trial_index: int,
    ) -> None:
        self.should_generate = should_generate
        self.reason = reason
        self.trial_index = trial_index

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> HealthcheckAnalysisCard:
        status = (
            HealthcheckStatus.PASS
            if self.should_generate
            else HealthcheckStatus.WARNING
        )
        return create_healthcheck_analysis_card(
            name=self.__class__.__name__,
            title=f"Ready to Generate Candidates for Trial {self.trial_index}",
            subtitle=self.reason,
            df=pd.DataFrame(),
            status=status,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import random
from unittest.mock import patch

import pandas as pd
from ax.core.data import Data
from ax.core.map_data import MAP_KEY, MapData
from ax.utils.common.testutils import TestCase
from ax.utils.common.timeutils import current_timestamp_in_millis
from pyre_extensions import assert_is_instance

REPR_1000: str = (
    "Data(df=\n|    |   trial_index |   arm_name | metric_name   |   mean |   sem "
    "| start_time          | end_time            |\n"
    "|---:|--------------:|-----------:|:--------------|-------:|------:"
    "|:--------------------|:--------------------|\n"
    "|  0 |             1 |        0_0 | a             |    2   |   0.2 "
    "| 2018-01-01 00:00:00 | 2018-01-02 00:00:00 |\n"
    "|  1 |             1 |        0_0 | b             |    1.8 |   0.3 "
    "| 2018-01-01 00:00:00 | 2018-01-02 00:00:00 |\n"
    "|  2 |             1 |        0_1 | a             |    4   |   0.6 "
    "| 2018-01-01 00:00:00 | 2018-01-02 00:00:00 |\n"
    "|  3 |             1 |        0_1 | b             |    3.7 |   0.5 "
    "| 2018-01-01 00:00:00 | 2018-01-02 00:00:00 |\n"
    "|  4 |             1 |        0_2 | a             |    0.5 | nan   "
    "| 2018-01-01 00:00:00 | 2018-01-02 00:00:00 |\n"
    "|  5 |             1 |        0_2 | b             |    3   | nan   "
    "| 2018-01-01 00:00:00 | 2018-01-02 00:00:00 |)"
)

REPR_500: str = (
    "Data(df=\n|    |   trial_index |   arm_name | metric_name   |   mean |   sem "
    "| start_time          | end_time            |\n"
    "|---:|--------------:|-----------:|:--------------|-------:|------:"
    "|:--------------------|:--------------------|\n"
    "|  0 |             1 |        0_0 | a             |    2   |   0.2 "
    "| 2018-01-01 00:00:00 | 2018-01-02 00:00:00 |\n"
    "|  1 |             1 |        0_0 | b             |    1.8 |   0.3 "
    "| 2018-01-01 00:00:00 | 2018-01-02 00:00:00 |\n"
    "|  2 |             1 |        0_1 | a           ...)"
)


def get_test_dataframe() -> pd.DataFrame:
    return pd.DataFrame(
        [
            {
                "arm_name": "0_0",
                "mean": 2.0,
                "sem": 0.2,
                "trial_index": 1,
                "metric_name": "a",
                "start_time": "2018-01-01",
                "end_time": "2018-01-02",
            },
            {
                "arm_name": "0_0",
                "mean": 1.8,
                "sem": 0.3,
                "trial_index": 1,
                "metric_name": "b",
                "start_time": "2018-01-01",
                "end_time": "2018-01-02",
            },
            {
                "arm_name": "0_1",
                "mean": 4.0,
                "sem": 0.6,
                "trial_index": 1,
                "metric_name": "a",
                "start_time": "2018-01-01",
                "end_time": "2018-01-02",
            },
            {
                "arm_name": "0_1",
                "mean": 3.7,
                "sem": 0.5,
                "trial_index": 1,
                "metric_name": "b",
                "start_time": "2018-01-01",
                "end_time": "2018-01-02",
            },
            {
                "arm_name": "0_2",
                "mean": 0.5,
                "sem": None,
                "trial_index": 1,
                "metric_name": "a",
                "start_time": "2018-01-01",
                "end_time": "2018-01-02",
            },
            {
                "arm_name": "0_2",
                "mean": 3.0,
                "sem": None,
                "trial_index": 1,
                "metric_name": "b",
                "start_time": "2018-01-01",
                "end_time": "2018-01-02",
            },
        ]
    )


class TestDataBase(TestCase):
    """
    Covers both Data and MapData tests.

    MapData tests are in test_map_data.py.
    """

    cls: type[Data] = Data

    def setUp(self) -> None:
        super().setUp()
        df = get_test_dataframe()

        if self.cls is Data:
            self.df = df
            self.data_with_df = Data(df=self.df)
            self.data_without_df = Data()
        else:
            df_1 = df.copy().assign(**{MAP_KEY: 0})
            df_2 = df.copy().assign(**{MAP_KEY: 1})
            self.df = pd.concat((df_1, df_2))
            self.data_with_df = MapData(df=self.df)
            self.data_without_df = MapData()

    def test_init(self) -> None:
        # For Data, this is Data(). For MapData, this is MapData(map_keys=[]).
        self.assertEqual(self.data_without_df, self.data_without_df)
        self.assertEqual(self.data_with_df, self.data_with_df)

        df = self.data_with_df.df
        self.assertEqual(
            float(df[df["arm_name"] == "0_0"][df["metric_name"] == "a"]["mean"].item()),
            2.0,
        )
        self.assertEqual(
            float(df[df["arm_name"] == "0_1"][df["metric_name"] == "b"]["sem"].item()),
            0.5,
        )

    def test_clone(self) -> None:
        data = self.data_with_df
        data._db_id = 1234
        data_clone = data.clone()
        # Check equality of the objects.
        self.assertTrue(data.df.equals(data_clone.df))
        # Make sure it's not the original object or df.
        self.assertIsNot(data, data_clone)
        self.assertIsNot(data.df, data_clone.df)
        self.assertIsNone(data_clone._db_id)
        if self.cls is MapData:
            data = assert_is_instance(data, MapData)
            data_clone = assert_is_instance(data_clone, MapData)
            self.assertIsNot(data.map_df, data_clone.map_df)
            self.assertTrue(data.map_df.equals(data_clone.map_df))

    def test_BadData(self) -> None:
        df = pd.DataFrame([{"bad_field": "0_0", "bad_field_2": {"x": 0, "y": "a"}}])
        with self.assertRaisesRegex(
            ValueError, "Dataframe must contain required columns"
        ):
            if self.cls is Data:
                Data(df=df)
            else:
                MapData(df=df)

    def test_EmptyData(self) -> None:
        data = self.data_without_df
        df = data.df
        self.assertTrue(df.empty)
        self.assertTrue(self.cls.from_multiple_data([]).df.empty)

        if isinstance(data, MapData):
            self.assertTrue(data.map_df.empty)
            expected_columns = Data.REQUIRED_COLUMNS.union({MAP_KEY})
        else:
            expected_columns = Data.REQUIRED_COLUMNS
        self.assertEqual(expected_columns, data.required_columns())
        self.assertEqual(set(df.columns), expected_columns)

    def test_from_multiple_with_generator(self) -> None:
        data = self.cls.from_multiple_data(self.data_with_df for _ in range(2))
        self.assertEqual(len(data.true_df), 2 * len(self.data_with_df.true_df))

    def test_data_column_data_types_default(self) -> None:
        self.assertEqual(self.cls.column_data_types(), self.cls.COLUMN_DATA_TYPES)

    def test_data_column_data_types_with_extra_columns(self) -> None:
        bartype = random.choice([str, int, float])
        columns = self.cls.column_data_types(extra_column_types={"foo": bartype})
        for c, t in self.cls.COLUMN_DATA_TYPES.items():
            self.assertEqual(columns[c], t)
        self.assertEqual(columns["foo"], bartype)


class DataTest(TestCase):
    """Tests that are specific to Data and not shared with MapData."""

    def setUp(self) -> None:
        super().setUp()
        self.df_hash = "be6ca1edb2d83e08c460665476d32caa"
        self.df = get_test_dataframe()

    def test_repr(self) -> None:
        self.assertEqual(
            str(Data(df=self.df)),
            REPR_1000,
        )
        with patch(f"{Data.__module__}.DF_REPR_MAX_LENGTH", 500):
            self.assertEqual(str(Data(df=self.df)), REPR_500)

    def test_OtherClassInequality(self) -> None:
        class CustomData(Data):
            pass

        data = CustomData(df=self.df)
        self.assertNotEqual(data, Data(self.df))

        # Try making regular data with extra column
        with self.assertRaisesRegex(ValueError, "cat"):
            Data(df=self.df.assign(cat="dog"))

    def test_FromEvaluationsIsoFormat(self) -> None:
        now = pd.Timestamp.now()
        day = now.day
        for sem in (0.5, None):
            eval1 = (3.7, sem) if sem is not None else 3.7
            data = Data.from_evaluations(
                evaluations={"0_1": {"b": eval1}},
                trial_index=0,
                sample_sizes={"0_1": 2},
                start_time=now.isoformat(),
                end_time=now.isoformat(),
            )
            self.assertEqual(data.df["sem"].isnull()[0], sem is None)
            self.assertEqual(len(data.df), 1)
            self.assertNotEqual(data, Data(self.df))
            self.assertEqual(data.df["start_time"][0].day, day)
            self.assertEqual(data.df["end_time"][0].day, day)

    def test_FromEvaluationsMillisecondFormat(self) -> None:
        now_ms = current_timestamp_in_millis()
        day = pd.Timestamp(now_ms, unit="ms").day
        for sem in (0.5, None):
            eval1 = (3.7, sem) if sem is not None else 3.7
            data = Data.from_evaluations(
                evaluations={"0_1": {"b": eval1}},
                trial_index=0,
                sample_sizes={"0_1": 2},
                start_time=now_ms,
                end_time=now_ms,
            )
            self.assertEqual(data.df["sem"].isnull()[0], sem is None)
            self.assertEqual(len(data.df), 1)
            self.assertNotEqual(data, Data(self.df))
            self.assertEqual(data.df["start_time"][0].day, day)
            self.assertEqual(data.df["end_time"][0].day, day)

    def test_from_multiple(self) -> None:
        with self.subTest("Combinining non-empty Data"):
            data = Data.from_multiple_data([Data(df=self.df), Data(df=self.df)])
            self.assertEqual(len(data.df), 2 * len(self.df))

        with self.subTest("Combining empty Data makes empty Data"):
            data = Data.from_multiple_data([Data(), Data()])
            self.assertEqual(data, Data())

        with self.subTest("Can't combine different types"):

            class CustomData(Data):
                pass

            with self.assertRaisesRegex(
                TypeError, "All data objects must be instances of"
            ):
                CustomData.from_multiple_data([Data(), CustomData()])

    def test_FromMultipleDataMismatchedTypes(self) -> None:
        # create two custom data types
        class CustomDataA(Data):
            pass

        class CustomDataB(Data):
            pass

        # Test using `Data.from_multiple_data` to combine non-Data types
        with self.assertRaisesRegex(TypeError, "All data objects must be instances of"):
            Data.from_multiple_data([CustomDataA(), CustomDataB()])

        # Test data of multiple non-empty types raises a value error
        data_elt_A = CustomDataA(df=self.df)
        data_elt_B = CustomDataB(df=self.df)
        with self.assertRaisesRegex(TypeError, "All data objects must be instances of"):
            Data.from_multiple_data([data_elt_A, data_elt_B])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Union

import numpy as np
import pandas as pd
from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis

from ax.analysis.healthcheck.healthcheck_analysis import (
    create_healthcheck_analysis_card,
    HealthcheckAnalysisCard,
    HealthcheckStatus,
)
from ax.core.experiment import Experiment

from ax.core.parameter import ChoiceParameter, Parameter, RangeParameter
from ax.core.parameter_constraint import ParameterConstraint
from ax.core.search_space import SearchSpace
from ax.core.types import TParameterization
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import assert_is_instance, override

SUBTITLE_BASE = (
    "This optimization may benefit from widened search-space bounds. Consider the "
    "expansions recommended below.\n"
)


class SearchSpaceAnalysis(Analysis):
    r"""
    Analysis for checking wehther the search space of the experiment should be expanded.
    It checks whether the suggested parameters land at the boundary of the search space
    and recommends expanding the search space if the proportion of the suggested
    parameters that land at the boundary is above the threshold.
    """

    def __init__(
        self, trial_index: int, boundary_proportion_threshold: float = 0.5
    ) -> None:
        r"""
        Args:
            trial_index: The index of the trial to analyze.
            boundary_proportion_threshold: The threshold on the proportion of suggested
                candidates that land on the boundary of the search space for us
                to recommend expanding the search space.

        Returns None
        """
        self.trial_index = trial_index
        self.boundary_proportion_threshold = boundary_proportion_threshold

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> HealthcheckAnalysisCard:
        if experiment is None:
            raise UserInputError("SearchSpaceAnalysis requires an Experiment.")

        status = HealthcheckStatus.PASS
        title_status = "Success"

        trial = experiment.trials[self.trial_index]
        arms = trial.arms
        parametrizations = [arm.parameters for arm in arms]
        boundary_proportions_df = search_space_boundary_proportions(
            search_space=experiment.search_space,
            parameterizations=parametrizations,
        )
        if np.any(
            boundary_proportions_df["proportion"] > self.boundary_proportion_threshold
        ):
            msg = boundary_proportions_message(
                boundary_proportions_df=boundary_proportions_df,
                boundary_proportion_threshold=self.boundary_proportion_threshold,
            )
            status = HealthcheckStatus.WARNING
            additional_subtitle = msg
            title_status = "Warning"
        else:
            additional_subtitle = "Search space does not need to be updated."

        return create_healthcheck_analysis_card(
            name=self.__class__.__name__,
            title=f"Ax search-space boundary check [{title_status}]",
            subtitle=SUBTITLE_BASE + additional_subtitle,
            df=boundary_proportions_df[["boundary", "proportion", "bound"]],
            status=status,
        )


def search_space_boundary_proportions(
    search_space: SearchSpace,
    parameterizations: list[TParameterization],
    tol: float = 1e-6,
) -> pd.DataFrame:
    r"""
    Compute the fractions of parameterizations that landed at the parameter and
    parameter constraint boundaries of the search space.

    Args:
        search_space: Search space.
        parameterizations: A list of suggested parameterizations (parameter values).
        tol: Relative tolerance for the difference between parameters and the
            boundary bounds.

    Returns:
        A dataframe containing parameters along with the fractions of parameterizations
        that landed at the parameter lower and upper limit in case of range and ordered
        choice parameters and containing parameter constraints along with the fractions
        of parameterizations that landed at the constraint boundary.
    """

    parameters_and_constraints = []
    boundaries = []
    proportions = []
    bounds = []

    parameterizations = [
        parameterization
        for parameterization in parameterizations
        if search_space.check_membership(parameterization)
    ]
    num_parameterizations = len(parameterizations)

    for parameter_name, parameter in search_space.parameters.items():
        if isinstance(parameter, RangeParameter):
            lower = parameter.lower
            upper = parameter.upper
        elif isinstance(parameter, ChoiceParameter) and parameter.is_ordered:
            values = [
                assert_is_instance(v, Union[int, float]) for v in parameter.values
            ]
            lower = min(values)
            upper = max(values)
        else:
            continue
        num_lb = 0  # counts how many parameters are equal to the boundary's lower bound
        num_ub = 0  # counts how many parameters are equal to the boundary's upper bound
        for parameterization in parameterizations:
            value = parameterization[parameter_name]
            value = float(value)
            # for choice parameters, we check if the value is equal to the lower
            # or upper bound
            if isinstance(search_space.parameters[parameter_name], ChoiceParameter):
                num_lb += int(value == lower)
                num_ub += int(value == upper)
            else:
                # for range paramaters, we check if the value is within the tolerance
                if abs(value - float(lower)) < tol * (float(upper) - float(lower)):
                    num_lb += 1
                elif abs(value - float(upper)) < tol * (float(upper) - float(lower)):
                    num_ub += 1
        if num_parameterizations != 0:
            prop_lower = num_lb / float(num_parameterizations)
            prop_upper = num_ub / float(num_parameterizations)
        else:
            prop_lower = 0.0
            prop_upper = 0.0
        parameters_and_constraints.extend([parameter] * 2)
        boundaries.extend(
            [f"{parameter_name} = {lower}", f"{parameter_name} = {upper}"]
        )
        proportions.extend([prop_lower, prop_upper])
        bounds.extend(["lower", "upper"])
    for pc in search_space.parameter_constraints:
        weighted_sums = [
            sum(
                float(assert_is_instance(parameterization[param], Union[int, float]))
                * weight
                for param, weight in pc.constraint_dict.items()
            )
            for parameterization in parameterizations
        ]
        if num_parameterizations != 0:
            prop = (
                np.sum(
                    [
                        abs(weighted_sum - pc.bound) < tol
                        for weighted_sum in weighted_sums
                    ]
                )
                / num_parameterizations
            )
        else:
            prop = 0.0
        boundaries.append(
            " + ".join(f"{v}*{k}" for k, v in sorted(pc.constraint_dict.items()))
            + f" = {pc.bound}"
        )
        proportions.append(prop)
        parameters_and_constraints.append(pc)
        bounds.append("upper")

    df = pd.DataFrame(
        {
            "parameter_or_constraint": parameters_and_constraints,
            "boundary": boundaries,
            "proportion": proportions,
            "bound": bounds,
        }
    )
    return df


def boundary_proportions_message(
    boundary_proportions_df: pd.DataFrame,
    boundary_proportion_threshold: float = 0.5,
) -> str:
    r"""
    Construct a message explaning what parameter or parameter constraints bounds
    to change based on the proportions of the parametrizations that landed at
    the search spaces boundaries.
    A proportion should be above the threshold in order to recommend expanding
    the search space along the corresponding parameter or parameter constraint.

    Args:
        boundary_proportions_df: A dataframe with the following columns
            * parameter_or_constraint: the parameter or constraint object
                containing this row's search space boundary.
            * boundary: a string representation of the function defining this boundary.
            * proportion: the proportion of provided parameterizations within
                ``tol`` of this boundary.
            * bound: whether this is an upper or lower bound.
        A dataframe containing parameters and parameter
            constraints along with the proportions of the parametrizations that
            landed at the lower or upper bounds of the parameters or at the
            constraints boundary.
        boundary_proportion_threshold: The minimal proportion of suggested
            parametrizations that land at the boundary of the search space for
            us to recommend expanding the search space.

    Returns:
        A string explaning what parameter or parameter constraints bounds to change
        in order to expand the search space.
    """

    msg = ""
    for _, row in boundary_proportions_df.iterrows():
        if isinstance(row["parameter_or_constraint"], Parameter):
            parameter = row["parameter_or_constraint"]
            bound = row["bound"]
            prop = row["proportion"]
            boundary = row["boundary"]
            if prop >= boundary_proportion_threshold:
                change_dir = "decreasing" if bound == "lower" else "increasing"
                msg += (
                    f"\n - **Relax {bound} bound of `{parameter.name!r}`:** Ax is "
                    f"frequently suggesting values at the {bound} bound of "
                    f"`{parameter.name!r}`, `{boundary}`. This may indicate that the "
                    f"optimal value of this parameter is outside this bound, in which "
                    f"case {change_dir} this {bound} bound would improve optimization "
                    f"performance. Details: {prop * 100:.2f}% of suggested arms are on "
                    f"the parameter's {bound} bound (threshold for this alert is "
                    f"{boundary_proportion_threshold * 100:.2f}%)."
                )
        elif isinstance(row["parameter_or_constraint"], ParameterConstraint):
            pc = row["parameter_or_constraint"]
            prop = row["proportion"]
            if prop >= boundary_proportion_threshold:
                msg += (
                    f"\n - **Relax parameter constraint `{pc!r}`:** Ax is frequently "
                    "suggesting parameterizations along the boundary of parameter"
                    f"constraint `{pc!r}`. Consider relaxing this constraint bound "
                    "somewhat and re-generating the candidates. "
                    f"Details: {prop * 100:.2f}% of all suggested arms lie along the "
                    "boundary, which exceeds the threshold of "
                    f"{boundary_proportion_threshold * 100:.2f}%."
                )

    return msg



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.api.protocols.metric import IMetric
from ax.api.protocols.runner import IRunner

__all__ = [
    "IMetric",
    "IRunner",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.benchmark.problems.synthetic.bandit import get_bandit_problem, get_baseline
from ax.utils.common.testutils import TestCase


class TestProblems(TestCase):
    def test_get_baseline(self) -> None:
        num_choices = 5
        baseline = get_baseline(num_choices=num_choices, n_sims=100)
        self.assertGreater(baseline, 0)
        # Worst = num_choices - 1; random guessing = (num_choices - 1) / 2
        self.assertLess(baseline, (num_choices - 1) / 2)

    def test_get_bandit_problem(self) -> None:
        problem = get_bandit_problem()
        self.assertEqual(problem.name, "Bandit")
        self.assertEqual(problem.num_trials, 3)
        self.assertTrue(problem.report_inference_value_as_trace)

        problem = get_bandit_problem(num_choices=26, num_trials=4)
        self.assertEqual(problem.num_trials, 4)
        self.assertEqual(problem.status_quo_params, {"x0": 26 // 2})

    def test_baseline_exception(self) -> None:
        with self.assertWarnsRegex(
            Warning, expected_regex="Baseline value is not available for num_choices=17"
        ):
            problem = get_bandit_problem(num_choices=17)

        self.assertEqual(problem.baseline_value, get_bandit_problem().baseline_value)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import json
from itertools import product

from ax.adapter.registry import Generators
from ax.analysis.plotly.arm_effects import ArmEffectsPlot, compute_arm_effects_adhoc
from ax.analysis.plotly.plotly_analysis import PlotlyAnalysisCard
from ax.api.client import Client
from ax.api.configs import RangeParameterConfig
from ax.core.arm import Arm
from ax.core.metric import Metric
from ax.exceptions.core import AxError, UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_offline_experiments,
    get_online_experiments,
)
from ax.utils.testing.mock import mock_botorch_optimize
from ax.utils.testing.modeling_stubs import (
    get_default_generation_strategy_at_MBM_node,
    get_sobol_MBM_MTGP_gs,
)
from pyre_extensions import assert_is_instance, none_throws


class TestArmEffectsPlot(TestCase):
    @mock_botorch_optimize
    def setUp(self) -> None:
        super().setUp()

        self.client = Client()
        self.client.configure_experiment(
            name="test_experiment",
            parameters=[
                RangeParameterConfig(
                    name="x1",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
                RangeParameterConfig(
                    name="x2",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
            ],
        )
        self.client.configure_optimization(
            objective="foo", outcome_constraints=["bar >= -0.5"]
        )

        # Get two trials and fail one, giving us a ragged structure
        self.client.get_next_trials(max_trials=2)
        self.client.complete_trial(trial_index=0, raw_data={"foo": 1.0, "bar": 2.0})
        self.client.mark_trial_failed(trial_index=1)

        # Complete 5 trials successfully
        for _ in range(5):
            for trial_index, parameterization in self.client.get_next_trials(
                max_trials=1
            ).items():
                self.client.complete_trial(
                    trial_index=trial_index,
                    raw_data={
                        "foo": assert_is_instance(parameterization["x1"], float),
                        "bar": assert_is_instance(parameterization["x1"], float)
                        - 2 * assert_is_instance(parameterization["x2"], float),
                    },
                )

    def test_validation(self) -> None:
        with self.assertRaisesRegex(
            UserInputError, "Requested metrics .* are not present in the experiment."
        ):
            ArmEffectsPlot(metric_names=["foo", "bar", "baz"]).compute(
                experiment=self.client._experiment,
                generation_strategy=self.client._generation_strategy,
            )

        with self.assertRaisesRegex(
            UserInputError, "Trial with index .* not found in experiment."
        ):
            ArmEffectsPlot(trial_index=1998).compute(
                experiment=self.client._experiment,
                generation_strategy=self.client._generation_strategy,
            )

    def test_compute_raw(self) -> None:
        default_analysis = ArmEffectsPlot(use_model_predictions=False)

        cards = default_analysis.compute(
            experiment=self.client._experiment,
            generation_strategy=self.client._generation_strategy,
        ).flatten()

        # Check that we have cards for both metrics
        self.assertEqual(len(cards), 2)

        self.assertEqual(
            set(cards[0].df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "bar_mean",
                "bar_sem",
            },
        )

        self.assertEqual(
            set(cards[1].df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
            },
        )

        for card in cards:
            # Check that we have one row per arm and that each arm appears only once
            self.assertEqual(len(card.df), len(self.client._experiment.arms_by_name))
            for arm_name in self.client._experiment.arms_by_name:
                self.assertEqual((card.df["arm_name"] == arm_name).sum(), 1)

        # Check that all SEMs are NaN
        self.assertTrue(cards[0].df["bar_sem"].isna().all())
        self.assertTrue(cards[1].df["foo_sem"].isna().all())

    def test_compute_with_modeled(self) -> None:
        default_analysis = ArmEffectsPlot(use_model_predictions=True)

        cards = default_analysis.compute(
            experiment=self.client._experiment,
            generation_strategy=self.client._generation_strategy,
        ).flatten()

        # Check that we have cards for both metrics
        self.assertEqual(len(cards), 2)

        self.assertEqual(
            set(cards[0].df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "bar_mean",
                "bar_sem",
            },
        )

        self.assertEqual(
            set(cards[1].df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
            },
        )

        for card in cards:
            # Check that we have one row per arm and that each arm appears only once
            self.assertEqual(len(card.df), len(self.client._experiment.arms_by_name))
            for arm_name in self.client._experiment.arms_by_name:
                self.assertEqual((card.df["arm_name"] == arm_name).sum(), 1)

        # Check that all SEMs are not NaN
        self.assertFalse(cards[0].df["bar_sem"].isna().any())
        self.assertFalse(cards[1].df["foo_sem"].isna().any())

    def test_plot_p_feasible(self) -> None:
        default_analysis = ArmEffectsPlot(
            metric_names=["p_feasible", "foo"], use_model_predictions=True
        )

        cards = default_analysis.compute(
            experiment=self.client._experiment,
            generation_strategy=self.client._generation_strategy,
        ).flatten()

        # Check that we have cards for both metrics
        self.assertEqual(len(cards), 2)

        self.assertEqual(
            set(cards[0].df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
            },
        )
        self.assertEqual(
            set(cards[1].df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
            },
        )

        for card in cards:
            # Check that we have one row per arm and that each arm appears only once
            self.assertEqual(len(card.df), len(self.client._experiment.arms_by_name))
            for arm_name in self.client._experiment.arms_by_name:
                self.assertEqual((card.df["arm_name"] == arm_name).sum(), 1)

        # Check that all SEMs are not NaN
        self.assertTrue(cards[0].df["p_feasible_sem"].isna().all())
        self.assertFalse(cards[1].df["foo_sem"].isna().any())
        self.assertEqual(
            json.loads(cards[0].blob)["layout"]["yaxis"]["title"]["text"], "p_feasible"
        )

        # test that plot errors if p_feasible is a metric on the experiment
        self.client._experiment.add_tracking_metric(Metric(name="p_feasible"))
        with self.assertRaisesRegex(
            AxError, "p_feasible is reserved for plotting the probability"
        ):
            default_analysis.compute(
                experiment=self.client._experiment,
                generation_strategy=self.client._generation_strategy,
            )

    def test_compute_adhoc(self) -> None:
        # Use the same kwargs for typical and adhoc
        kwargs = {
            "metric_names": ["foo", "bar"],
            "use_model_predictions": True,
            "additional_arms": [Arm(parameters={"x1": 0, "x2": 0})],
            "labels": {"foo": "f"},
        }
        # pyre-ignore[6]: Unsafe kwargs usage on purpose
        analysis = ArmEffectsPlot(**kwargs)

        cards = analysis.compute(
            experiment=self.client._experiment,
            generation_strategy=self.client._generation_strategy,
        )

        adhoc_cards = compute_arm_effects_adhoc(
            experiment=self.client._experiment,
            generation_strategy=self.client._generation_strategy,
            # pyre-ignore[6]: Unsafe kwargs usage on purpose
            **kwargs,
        )

        self.assertEqual(cards, adhoc_cards)

    @TestCase.ax_long_test(
        reason=(
            "Adapter.predict still too slow under @mock_botorch_optimize for this test"
        )
    )
    @mock_botorch_optimize
    def test_online(self) -> None:
        # Test ArmEffectsPlot can be computed for a variety of experiments which
        # resemble those we see in an online setting.

        for experiment in get_online_experiments():
            arm = Generators.SOBOL(experiment=experiment).gen(n=1).arms[0]
            arm.name = "additional_arm"
            for (
                use_model_predictions,
                trial_index,
                with_additional_arms,
                show_cumulative_best,
            ) in product([True, False], [None, 0], [True, False], [True, False]):
                if use_model_predictions and with_additional_arms:
                    additional_arms = [arm]
                else:
                    additional_arms = None
                generation_strategy = get_default_generation_strategy_at_MBM_node(
                    experiment=experiment
                )
                generation_strategy.current_node._fit(experiment=experiment)
                adapter = none_throws(generation_strategy.adapter)

                analysis = ArmEffectsPlot(
                    metric_names=[*adapter.metric_names, "p_feasible"],
                    use_model_predictions=use_model_predictions,
                    trial_index=trial_index,
                    additional_arms=additional_arms,
                    show_cumulative_best=show_cumulative_best,
                )

                cards = analysis.compute(
                    experiment=experiment,
                    adapter=adapter,
                )
                self.assertEqual(len(cards.flatten()), len(adapter.metric_names) + 1)
                if with_additional_arms and use_model_predictions:
                    # validate that we plotted the additional arm
                    self.assertTrue(
                        any(
                            arm.name in datum["text"][0]
                            for card in cards.flatten()
                            for datum in assert_is_instance(card, PlotlyAnalysisCard)
                            .get_figure()
                            ._data
                        )
                    )

    @TestCase.ax_long_test(
        reason=(
            "Adapter.predict still too slow under @mock_botorch_optimize for this test"
        )
    )
    @mock_botorch_optimize
    def test_offline(self) -> None:
        # Test ArmEffectsPlot can be computed for a variety of experiments which
        # resemble those we see in an offline setting.

        for experiment in get_offline_experiments():
            for use_model_predictions in [True, False]:
                for trial_index in [None, 0]:
                    for with_additional_arms in [True, False]:
                        for show_cumulative_best in [True, False]:
                            if use_model_predictions and with_additional_arms:
                                additional_arms = [
                                    Arm(
                                        parameters={
                                            parameter_name: 0
                                            for parameter_name in (
                                                experiment.search_space.parameters.keys()  # noqa E501
                                            )
                                        }
                                    )
                                ]
                            else:
                                additional_arms = None

                            generation_strategy = (
                                get_default_generation_strategy_at_MBM_node(
                                    experiment=experiment
                                )
                            )
                            generation_strategy.current_node._fit(experiment=experiment)
                            adapter = none_throws(generation_strategy.adapter)

                            analysis = ArmEffectsPlot(
                                metric_names=[*adapter.metric_names],
                                use_model_predictions=use_model_predictions,
                                trial_index=trial_index,
                                additional_arms=additional_arms,
                                show_cumulative_best=show_cumulative_best,
                            )

                            _ = analysis.compute(
                                experiment=experiment,
                                adapter=adapter,
                            )


class TestArmEffectsPlotRel(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.experiment = get_branin_experiment(with_status_quo=True)
        self.generation_strategy = get_sobol_MBM_MTGP_gs()
        self.generation_strategy.experiment = self.experiment
        # Run 2 trials
        for _ in range(2):
            self.experiment.new_batch_trial(
                generator_runs=self.generation_strategy._gen_with_multiple_nodes(
                    experiment=self.experiment, n=3
                )
            ).add_status_quo_arm(weight=1.0).mark_completed(unsafe=True)
            self.experiment.fetch_data()

    def test_compute_with_relativize(self) -> None:
        for use_model_predictions in [True, False]:
            with self.subTest(use_model_predictions=use_model_predictions):
                analysis = ArmEffectsPlot(
                    use_model_predictions=use_model_predictions, relativize=True
                )

                cards = analysis.compute(
                    experiment=self.experiment,
                    generation_strategy=self.generation_strategy,
                ).flatten()

                self.assertEqual(len(cards), 1)

                self.assertEqual(
                    set(cards[0].df.columns),
                    {
                        "trial_index",
                        "arm_name",
                        "trial_status",
                        "generation_node",
                        "p_feasible_mean",
                        "p_feasible_sem",
                        "branin_mean",
                        "branin_sem",
                    },
                )

                for card in cards:
                    # Check that we have one row per arm and that each arm appears only
                    # once. Exclude status_quo since that is repeated between trials
                    card_arms = card.df[card.df.arm_name != "status_quo"].arm_name
                    experiment_arms = self.experiment.arms_by_name.copy()
                    experiment_arms.pop("status_quo")
                    self.assertEqual(len(card_arms), len(experiment_arms))
                    for arm_name in experiment_arms:
                        self.assertEqual((card.df["arm_name"] == arm_name).sum(), 1)

                    self.assertFalse(card.df["branin_mean"].isna().any())
                    self.assertFalse(card.df["branin_sem"].isna().any())



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import itertools
import warnings
from collections.abc import Callable, Mapping, Sequence
from typing import Protocol, TypeVar, Union

import numpy as np
import numpy.typing as npt
import torch
from ax.core.search_space import SearchSpaceDigest
from ax.exceptions.core import SearchSpaceExhausted, UnsupportedError
from ax.generators.types import TConfig
from botorch.acquisition.risk_measures import RiskMeasureMCObjective
from botorch.exceptions.warnings import OptimizationWarning
from pyre_extensions import assert_is_instance
from torch import Tensor


# pyre-fixme[24]: Generic type `np.ndarray` expects 2 type parameters.
Tensoray = Union[torch.Tensor, np.ndarray]
TTensoray = TypeVar("TTensoray", bound=Tensoray)


class TorchGeneratorLike(Protocol):
    """A protocol that stands in for ``TorchModel`` like objects that
    have a ``predict`` method.
    """

    def predict(
        self, X: Tensor, use_posterior_predictive: bool = False
    ) -> tuple[Tensor, Tensor]:
        """Predicts outcomes given an input tensor.

        Args:
            X: A ``n x d`` tensor of input parameters.
            use_posterior_predictive: A boolean indicating if the predictions
                should be from the posterior predictive (i.e. including
                observation noise).
                This option is only supported by the ``BoTorchGenerator``.

        Returns:
            Tensor: The predicted posterior mean as an ``n x o``-dim tensor.
            Tensor: The predicted posterior covariance as a ``n x o x o``-dim tensor.
        """
        ...


DEFAULT_MAX_RS_DRAWS = 10000


def rejection_sample(
    gen_unconstrained: Callable[
        [int, int, npt.NDArray, dict[int, float] | None], npt.NDArray
    ],
    n: int,
    d: int,
    tunable_feature_indices: npt.NDArray,
    linear_constraints: tuple[npt.NDArray, npt.NDArray] | None = None,
    deduplicate: bool = False,
    max_draws: int | None = None,
    fixed_features: dict[int, float] | None = None,
    rounding_func: Callable[[npt.NDArray], npt.NDArray] | None = None,
    existing_points: npt.NDArray | None = None,
) -> tuple[npt.NDArray, int]:
    """Rejection sample in parameter space. Parameter space is typically
    [0, 1] for all tunable parameters.

    Generators must implement a `gen_unconstrained` method in order to support
    rejection sampling via this utility.

    Args:
        gen_unconstrained: A callable that generates unconstrained points in
            the parameter space. This is typically the `_gen_unconstrained` method
            of a `RandomGenerator`.
        n: Number of samples to generate.
        d: Dimensionality of the parameter space.
        tunable_feature_indices: Indices of the tunable features in the
            parameter space.
        linear_constraints: A tuple of (A, b). For k linear constraints on
            d-dimensional x, A is (k x d) and b is (k x 1) such that
            A x <= b.
        deduplicate: If true, reject points that are duplicates of previously
            generated points. The points are deduplicated after applying the
            rounding function.
        max_draws: Maximum number of attemped draws before giving up.
        fixed_features: A map {feature_index: value} for features that
            should be fixed to a particular value during generation.
        rounding_func: A function that rounds an optimization result
            appropriately (e.g., according to `round-trip` transformations).
        existing_points: A set of previously generated points to use
            for deduplication. These should be provided in the parameter
            space model operates in.

    Returns:
        2-element tuple containing the generated points and the number of
        attempted draws.
    """
    # We need to perform the round trip transformation on our generated point
    # in order to deduplicate in the original search space.
    # The transformation is applied above.
    if deduplicate and rounding_func is None:
        raise ValueError("Rounding function must be provided for deduplication.")

    # Rejection sample with parameter constraints.
    points = np.zeros((n, d))
    attempted_draws = 0
    successful_draws = 0
    if max_draws is None:
        max_draws = DEFAULT_MAX_RS_DRAWS

    while successful_draws < n and attempted_draws <= max_draws:
        # _gen_unconstrained returns points including fixed features.
        # pyre-ignore: Anonymous function w/ named args.
        point = gen_unconstrained(
            n=1,
            d=d,
            tunable_feature_indices=tunable_feature_indices,
            fixed_features=fixed_features,
        )[0]

        # Check parameter constraints, always in raw transformed space.
        if linear_constraints is not None:
            all_constraints_satisfied, _ = check_param_constraints(
                linear_constraints=linear_constraints, point=point
            )
        else:
            all_constraints_satisfied = True

        if all_constraints_satisfied:
            # Apply the rounding function if the point satisfies the linear constraints.
            if rounding_func is not None:
                # NOTE: This could still fail rounding with a logger warning. But this
                # should be rare since the point is feasible in the continuous space.
                point = rounding_func(point)

            # Deduplicate: don't add the same point twice.
            duplicate = False
            if deduplicate:
                if existing_points is not None:
                    prev_points = np.vstack(
                        [points[:successful_draws, :], existing_points]
                    )
                else:
                    prev_points = points[:successful_draws, :]
                duplicate = check_duplicate(point=point, points=prev_points)

            # Add point if valid.
            if not duplicate:
                points[successful_draws] = point
                successful_draws += 1
        attempted_draws += 1

    if successful_draws < n:
        # Only possible if attempted_draws >= max_draws.
        raise SearchSpaceExhausted(
            f"Rejection sampling error (specified maximum draws ({max_draws}) exhausted"
            f", without finding sufficiently many ({n}) candidates). This likely means "
            "that there are no new points left in the search space."
        )

    return (points, attempted_draws)


def check_duplicate(point: npt.NDArray, points: npt.NDArray) -> bool:
    """Check if a point exists in another array.

    Args:
        point: Newly generated point to check.
        points: Points previously generated.

    Returns:
        True if the point is contained in points, else False
    """
    for p in points:
        if np.array_equal(p, point):
            return True
    return False


def add_fixed_features(
    tunable_points: npt.NDArray,
    d: int,
    fixed_features: dict[int, float] | None,
    tunable_feature_indices: npt.NDArray,
) -> npt.NDArray:
    """Add fixed features to points in tunable space.

    Args:
        tunable_points: Points in tunable space.
        d: Dimension of parameter space.
        fixed_features: A map {feature_index: value} for features that
            should be fixed to a particular value during generation.
        tunable_feature_indices: Parameter indices (in d) which are tunable.

    Returns:
        points: Points in the full d-dimensional space, defined by bounds.
    """
    n = np.shape(tunable_points)[0]
    points = np.zeros((n, d))
    points[:, tunable_feature_indices] = tunable_points
    if fixed_features:
        fixed_feature_indices = np.array(list(fixed_features.keys()))
        fixed_values = np.tile(list(fixed_features.values()), (n, 1))
        points[:, fixed_feature_indices] = fixed_values
    return points


def check_param_constraints(
    linear_constraints: tuple[npt.NDArray, npt.NDArray],
    point: npt.NDArray,
) -> tuple[bool, npt.NDArray]:
    """Check if a point satisfies parameter constraints.

    Args:
        linear_constraints: A tuple of (A, b). For k linear constraints on
            d-dimensional x, A is (k x d) and b is (k x 1) such that
            A x <= b.
        point: A candidate point in d-dimensional space, as a (1 x d) matrix.

    Returns:
        2-element tuple containing

        - Flag that is True if all constraints are satisfied by the point.
        - Indices of constraints which are violated by the point.
    """
    constraints_satisfied = (
        linear_constraints[0] @ np.expand_dims(point, axis=1) <= linear_constraints[1]
    )
    if np.all(constraints_satisfied):
        return True, np.array([])
    else:
        return (False, np.where(constraints_satisfied == False)[0])  # noqa: E712


def tunable_feature_indices(
    bounds: list[tuple[float, float]],
    fixed_features: dict[int, float] | None = None,
) -> npt.NDArray:
    """Get the feature indices of tunable features.

    Args:
        bounds: A list of (lower, upper) tuples for each column of X.
        fixed_features: A map {feature_index: value} for features that
            should be fixed to a particular value during generation.

    Returns:
        The indices of tunable features.
    """
    fixed_feature_indices = list(fixed_features.keys()) if fixed_features else []
    feature_indices = np.arange(len(bounds))
    return np.delete(feature_indices, fixed_feature_indices)


def validate_bounds(
    bounds: Sequence[tuple[float, float]],
    fixed_feature_indices: npt.NDArray,
) -> None:
    """Ensure the requested space is [0,1]^d.

    Args:
        bounds: A list of d (lower, upper) tuples for each column of X.
        fixed_feature_indices: Indices of features which are fixed at a
            particular value.
    """
    for feature_idx, bound in enumerate(bounds):
        # Bounds for fixed features are not unit-transformed.
        if feature_idx in fixed_feature_indices:
            continue

        if bound[0] != 0 or bound[1] != 1:
            raise ValueError(
                "This generator operates on [0,1]^d. Please make use "
                "of the UnitX transform in the Adapter, and ensure "
                "task features are fixed."
            )


def best_observed_point(
    model: TorchGeneratorLike,
    bounds: Sequence[tuple[float, float]],
    objective_weights: TTensoray | None,
    outcome_constraints: tuple[TTensoray, TTensoray] | None = None,
    linear_constraints: tuple[TTensoray, TTensoray] | None = None,
    fixed_features: dict[int, float] | None = None,
    risk_measure: RiskMeasureMCObjective | None = None,
    options: TConfig | None = None,
) -> TTensoray | None:
    """Select the best point that has been observed.

    Implements two approaches to selecting the best point.

    For both approaches, only points that satisfy parameter space constraints
    (bounds, linear_constraints, fixed_features) will be returned. Points must
    also be observed for all objective and constraint outcomes. Returned
    points may violate outcome constraints, depending on the method below.

    1: Select the point that maximizes the expected utility
    (objective_weights^T posterior_objective_means - baseline) * Prob(feasible)
    Here baseline should be selected so that at least one point has positive
    utility. It can be specified in the options dict, otherwise
    min (objective_weights^T posterior_objective_means)
    will be used, where the min is over observed points.

    2: Select the best-objective point that is feasible with at least
    probability p.

    The following quantities may be specified in the options dict:

    - best_point_method: 'max_utility' (default) or 'feasible_threshold'
      to select between the two approaches described above.
    - utility_baseline: Value for the baseline used in max_utility approach. If
      not provided, defaults to min objective value.
    - probability_threshold: Threshold for the feasible_threshold approach.
      Defaults to p=0.95.
    - feasibility_mc_samples: Number of MC samples used for estimating the
      probability of feasibility (defaults 10k).

    Args:
        model: A Torch model or Surrogate.
        bounds: A list of (lower, upper) tuples for each feature.
        objective_weights: The objective is to maximize a weighted sum of
            the columns of f(x). These are the weights.
        outcome_constraints: A tuple of (A, b). For k outcome constraints
            and m outputs at f(x), A is (k x m) and b is (k x 1) such that
            A f(x) <= b.
        linear_constraints: A tuple of (A, b). For k linear constraints on
            d-dimensional x, A is (k x d) and b is (k x 1) such that
            A x <= b.
        fixed_features: A map {feature_index: value} for features that
            should be fixed to a particular value in the best point.
        risk_measure: An optional risk measure for reporting best robust point.
        options: A config dictionary with settings described above.

    Returns:
        A d-array of the best point, or None if no feasible point exists.
    """
    if not hasattr(model, "Xs"):
        raise ValueError(f"Model must store training data Xs, but {model} does not.")
    best_point_and_value = best_in_sample_point(
        Xs=model.Xs,  # pyre-ignore[16]: Presence of attr. checked above.
        model=model,
        bounds=bounds,
        objective_weights=objective_weights,
        outcome_constraints=outcome_constraints,
        linear_constraints=linear_constraints,
        fixed_features=fixed_features,
        risk_measure=risk_measure,
        options=options,
    )
    return None if best_point_and_value is None else best_point_and_value[0]


def best_in_sample_point(
    Xs: Sequence[TTensoray],
    model: TorchGeneratorLike,
    bounds: Sequence[tuple[float, float]],
    objective_weights: TTensoray | None,
    outcome_constraints: tuple[TTensoray, TTensoray] | None = None,
    linear_constraints: tuple[TTensoray, TTensoray] | None = None,
    fixed_features: dict[int, float] | None = None,
    risk_measure: RiskMeasureMCObjective | None = None,
    options: TConfig | None = None,
) -> tuple[TTensoray, float] | None:
    """Select the best point that has been observed.

    Implements two approaches to selecting the best point.

    For both approaches, only points that satisfy parameter space constraints
    (bounds, linear_constraints, fixed_features) will be returned. Points must
    also be observed for all objective and constraint outcomes. Returned
    points may violate outcome constraints, depending on the method below.

    1: Select the point that maximizes the expected utility
    (objective_weights^T posterior_objective_means - baseline) * Prob(feasible)
    Here baseline should be selected so that at least one point has positive
    utility. It can be specified in the options dict, otherwise
    min (objective_weights^T posterior_objective_means)
    will be used, where the min is over observed points.

    2: Select the best-objective point that is feasible with at least
    probability p.

    The following quantities may be specified in the options dict:

    - best_point_method: 'max_utility' (default) or 'feasible_threshold'
      to select between the two approaches described above.
    - utility_baseline: Value for the baseline used in max_utility approach. If
      not provided, defaults to min objective value.
    - probability_threshold: Threshold for the feasible_threshold approach.
      Defaults to p=0.95.
    - feasibility_mc_samples: Number of MC samples used for estimating the
      probability of feasibility (defaults 10k).

    Args:
        Xs: Training data for the points, among which to select the best.
        model: A Torch model or Surrogate.
        bounds: A list of (lower, upper) tuples for each feature.
        objective_weights: The objective is to maximize a weighted sum of
            the columns of f(x). These are the weights.
        outcome_constraints: A tuple of (A, b). For k outcome constraints
            and m outputs at f(x), A is (k x m) and b is (k x 1) such that
            A f(x) <= b.
        linear_constraints: A tuple of (A, b). For k linear constraints on
            d-dimensional x, A is (k x d) and b is (k x 1) such that
            A x <= b.
        fixed_features: A map {feature_index: value} for features that
            should be fixed to a particular value in the best point.
        risk_measure: An optional risk measure for reporting best robust point.
        options: A config dictionary with settings described above.

    Returns:
        A two-element tuple or None if no feasible point exist. In tuple:
        - d-array of the best point,
        - utility at the best point.
    """
    if risk_measure is not None:
        # TODO[T131759268]: We need to apply the risk measure. Instead of doing obj_w @,
        # we could use `get_botorch_objective_and_transform` to get the objective
        # then apply it, though we also need to decide how to deal with constraints.
        raise NotImplementedError
    # Parse options
    if options is None:
        options = {}
    method: str = options.get("best_point_method", "max_utility")
    B: float | None = options.get("utility_baseline", None)
    threshold: float = options.get("probability_threshold", 0.95)
    nsamp: int = options.get("feasibility_mc_samples", 10000)
    # Get points observed for all objective and constraint outcomes
    if objective_weights is None:
        return None
    objective_weights_np = assert_is_instance(as_array(objective_weights), np.ndarray)
    X_obs = get_observed(
        Xs=Xs,
        objective_weights=objective_weights,
        outcome_constraints=outcome_constraints,
    )
    # Filter to those that satisfy constraints.
    X_obs = filter_constraints_and_fixed_features(
        X=X_obs,
        bounds=bounds,
        linear_constraints=linear_constraints,
        fixed_features=fixed_features,
    )
    if len(X_obs) == 0:
        # No feasible points
        return None
    # Predict objective and P(feas) at these points for Torch models.
    if isinstance(Xs[0], torch.Tensor):
        # pyre-fixme[16]: Item `ndarray` of `Union[ndarray[typing.Any, typing.Any],
        #  Tensor]` has no attribute `detach`.
        X_obs = X_obs.detach().clone()
    # (n_feasible x n_outcomes), (n_feasible x n_outcomes x n_outcomes)
    f, cov = as_array(model.predict(X_obs))
    # (n_outcomes,) x (n_outcomes, n_feasible) => (n_feasible,)
    obj = objective_weights_np @ f.transpose()
    pfeas = np.ones_like(obj)
    if outcome_constraints is not None:
        # (n_constraints x n_outcomes) and (n_constraints x 1)
        A, b = as_array(outcome_constraints)
        # Use Monte Carlo to compute pfeas, to properly handle covariance
        # across outcomes.
        for i, _ in enumerate(X_obs):
            # nsamp x n_outcomes
            z = np.random.multivariate_normal(
                mean=f[i, :], cov=cov[i, :, :], size=nsamp
            )
            # (n_constraints x n_outcomes) @ (n_outcomes x nsamp)
            pfeas[i] = (A @ z.transpose() <= b).all(axis=0).mean()
    # Identify best point
    if method == "feasible_threshold":
        utility = obj
        utility[pfeas < threshold] = -np.inf
    elif method == "max_utility":
        if B is None:
            B = obj.min()
        utility = (obj - B) * pfeas
    else:  # pragma: no cover
        raise UnsupportedError(f"Unknown best point method {method}.")
    i = np.argmax(utility)
    if utility[i] == -np.inf:
        return None
    else:
        return X_obs[i, :], utility[i]


def as_array(
    x: Tensoray | tuple[Tensoray, ...],
) -> npt.NDArray | tuple[npt.NDArray, ...]:
    """Convert every item in a tuple of tensors/arrays into an array.

    Args:
        x: A tensor, array, or a tuple of potentially mixed tensors and arrays.

    Returns:
        x, with everything converted to array.
    """
    if isinstance(x, tuple):
        return tuple(as_array(x_i) for x_i in x)  # pyre-ignore
    elif isinstance(x, np.ndarray):
        return x
    elif torch.is_tensor(x):
        return x.detach().cpu().double().numpy()
    else:
        raise ValueError("Input to as_array must be numpy array or torch tensor")


def get_observed(
    Xs: Sequence[TTensoray],
    objective_weights: TTensoray,
    outcome_constraints: tuple[TTensoray, TTensoray] | None = None,
) -> TTensoray:
    """Filter points to those that are observed for objective outcomes and outcomes
    that show up in outcome_constraints (if there are any).

    Args:
        Xs: A list of m (k_i x d) feature matrices X. Number of rows k_i
            can vary from i=1,...,m.
        objective_weights: The objective is to maximize a weighted sum of
            the columns of f(x). These are the weights.
        outcome_constraints: A tuple of (A, b). For k outcome constraints
            and m outputs at f(x), A is (k x m) and b is (k x 1) such that
            A f(x) <= b.

    Returns:
        Points observed for all objective outcomes and outcome constraints.
    """
    objective_weights_np = as_array(objective_weights)
    used_outcomes: set[int] = set(np.where(objective_weights_np != 0)[0])
    if len(used_outcomes) == 0:
        raise ValueError("At least one objective weight must be non-zero")
    if outcome_constraints is not None:
        used_outcomes = used_outcomes.union(
            np.where(as_array(outcome_constraints)[0] != 0)[1]
        )
    outcome_list = list(used_outcomes)
    X_obs_set = {tuple(float(x_i) for x_i in x) for x in Xs[outcome_list[0]]}
    for _, idx in enumerate(outcome_list, start=1):
        X_obs_set = X_obs_set.intersection(
            {tuple(float(x_i) for x_i in x) for x in Xs[idx]}
        )
    if isinstance(Xs[0], np.ndarray):
        # pyre-fixme[7]: This function only returns a Numpy array when Xs
        # contains all Numpy arrays, but Pyre doesn't understand
        return np.array(list(X_obs_set), dtype=Xs[0].dtype)  # (n x d)
    # pyre-fixme[7]: This function only returns a tensor when Xs
    # contains all tensors, but Pyre doesn't understand`.
    return torch.tensor(
        list(X_obs_set),
        device=assert_is_instance(Xs[0], torch.Tensor).device,
        dtype=Xs[0].dtype,
    )


def filter_constraints_and_fixed_features(
    X: TTensoray,
    bounds: Sequence[tuple[float, float]],
    linear_constraints: tuple[TTensoray, TTensoray] | None = None,
    fixed_features: dict[int, float] | None = None,
) -> TTensoray:
    """Filter points to those that satisfy bounds, linear_constraints, and
    fixed_features.

    Args:
        X: An tensor or array of points.
        bounds: A list of (lower, upper) tuples for each feature.
        linear_constraints: A tuple of (A, b). For k linear constraints on
            d-dimensional x, A is (k x d) and b is (k x 1) such that
            A x <= b.
        fixed_features: A map {feature_index: value} for features that
            should be fixed to a particular value in the best point.

    Returns:
        Feasible points.
    """
    if len(X) == 0:  # if there are no points, nothing to filter
        return X
    # pyre-ignore: Undefined attribute [16]: `np.ndarray` has no attribute
    # `cpu`.
    X_np = X.cpu().numpy() if isinstance(X, torch.Tensor) else X
    feas = np.ones(X_np.shape[0], dtype=bool)  # (n)
    for i, b in enumerate(bounds):
        feas &= (X_np[:, i] >= b[0]) & (X_np[:, i] <= b[1])
    if linear_constraints is not None:
        A, b = as_array(linear_constraints)  # (m x d) and (m x 1)
        feas &= (A @ X_np.transpose() <= b).all(axis=0)
    if fixed_features is not None:
        for idx, val in fixed_features.items():
            feas &= X_np[:, idx] == val
    X_feas = X_np[feas, :]
    if isinstance(X, torch.Tensor):
        return torch.from_numpy(X_feas).to(device=X.device, dtype=X.dtype)
    return X_feas


def mk_discrete_choices(
    ssd: SearchSpaceDigest,
    fixed_features: Mapping[int, float] | None = None,
) -> Mapping[int, Sequence[float]]:
    discrete_choices = ssd.discrete_choices
    # Add in fixed features.
    if fixed_features is not None:
        # Note: if any discrete features are fixed we won't enumerate those.
        discrete_choices = {
            **discrete_choices,
            **{k: [v] for k, v in fixed_features.items()},
        }
    return discrete_choices


def enumerate_discrete_combinations(
    discrete_choices: Mapping[int, Sequence[float]],
) -> list[dict[int, float]]:
    n_combos = np.prod([len(v) for v in discrete_choices.values()])
    if n_combos > 50:
        warnings.warn(
            f"Enumerating {n_combos} combinations of discrete parameter values "
            "while optimizing over a mixed search space. This can be very slow.",
            OptimizationWarning,
            stacklevel=2,
        )
    fixed_features_list = [
        dict(zip(discrete_choices.keys(), c))
        for c in itertools.product(*discrete_choices.values())
    ]
    return fixed_features_list


def all_ordinal_features_are_integer_valued(
    ssd: SearchSpaceDigest,
) -> bool:
    """Check if all ordinal features are integer-valued.

    Args:
        ssd: A SearchSpaceDigest.

    Returns:
        True if all ordinal features are integer-valued, False otherwise.
    """
    for feature_idx in ssd.ordinal_features:
        choices = ssd.discrete_choices[feature_idx]
        int_choices = [int(c) for c in choices]
        if choices != int_choices:
            return False
    return True



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import warnings
from collections.abc import Callable, Sequence
from copy import deepcopy
from typing import Any, Optional

import numpy.typing as npt
import torch
from ax.core.search_space import SearchSpaceDigest
from ax.core.types import TCandidateMetadata
from ax.exceptions.core import AxWarning, DataRequiredError
from ax.generators.torch.botorch_defaults import (
    get_and_fit_model,
    get_qLogNEI,
    recommend_best_observed_point,
    scipy_optimizer,
    TAcqfConstructor,
)
from ax.generators.torch.utils import (
    _datasets_to_legacy_inputs,
    _get_X_pending_and_observed,
    _to_inequality_constraints,
    predict_from_model,
    subset_model,
)
from ax.generators.torch_base import TorchGenerator, TorchGenResults, TorchOptConfig
from ax.generators.types import TConfig
from ax.utils.common.constants import Keys
from ax.utils.common.docutils import copy_doc
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.models import ModelList
from botorch.models.model import Model
from botorch.utils.datasets import SupervisedDataset
from botorch.utils.transforms import is_ensemble, normalize_indices
from pyre_extensions import assert_is_instance, none_throws
from torch import Tensor
from torch.nn import ModuleList  # @manual


# pyre-fixme[33]: Aliased annotation cannot contain `Any`.
TModelConstructor = Callable[
    [
        list[Tensor],
        list[Tensor],
        list[Tensor],
        list[int],
        list[int],
        list[str],
        Optional[dict[str, Tensor]],
        Any,
    ],
    Model,
]
TModelPredictor = Callable[[Model, Tensor, bool], tuple[Tensor, Tensor]]


# pyre-fixme[33]: Aliased annotation cannot contain `Any`.
TOptimizer = Callable[
    [
        AcquisitionFunction,
        Tensor,
        int,
        Optional[list[tuple[Tensor, Tensor, float]]],
        Optional[list[tuple[Tensor, Tensor, float]]],
        Optional[dict[int, float]],
        Optional[Callable[[Tensor], Tensor]],
        Any,
    ],
    tuple[Tensor, Tensor],
]
TBestPointRecommender = Callable[
    [
        TorchGenerator,
        list[tuple[float, float]],
        Tensor,
        Optional[tuple[Tensor, Tensor]],
        Optional[tuple[Tensor, Tensor]],
        Optional[dict[int, float]],
        Optional[TConfig],
        Optional[dict[int, float]],
    ],
    Optional[Tensor],
]


class LegacyBoTorchGenerator(TorchGenerator):
    r"""
    Customizable botorch model.

    By default, this uses a noisy Log Expected Improvement (qLogNEI) acquisition
    function on top of a model made up of separate GPs, one for each outcome. This
    behavior can be modified by providing custom implementations of the following
    components:

    - a `model_constructor` that instantiates and fits a model on data
    - a `model_predictor` that predicts outcomes using the fitted model
    - a `acqf_constructor` that creates an acquisition function from a fitted model
    - a `acqf_optimizer` that optimizes the acquisition function
    - a `best_point_recommender` that recommends a current "best" point (i.e.,
        what the model recommends if the learning process ended now)

    Args:
        model_constructor: A callable that instantiates and fits a model on data,
            with signature as described below.
        model_predictor: A callable that predicts using the fitted model, with
            signature as described below.
        acqf_constructor: A callable that creates an acquisition function from a
            fitted model, with signature as described below.
        acqf_optimizer: A callable that optimizes the acquisition function, with
            signature as described below.
        best_point_recommender: A callable that recommends the best point, with
            signature as described below.
        refit_on_cv: If True, refit the model for each fold when performing
            cross-validation.
        warm_start_refitting: If True, start model refitting from previous
            model parameters in order to speed up the fitting process.
        prior: An optional dictionary that contains the specification of GP model prior.
            Currently, the keys include:
            - covar_module_prior: prior on covariance matrix e.g.
                {"lengthscale_prior": GammaPrior(3.0, 6.0)}.
            - type: type of prior on task covariance matrix e.g.`LKJCovariancePrior`.
            - sd_prior: A scalar prior over nonnegative numbers, which is used for the
                default LKJCovariancePrior task_covar_prior.
            - eta: The eta parameter on the default LKJ task_covar_prior.


    Call signatures:

    ::

        model_constructor(
            Xs,
            Ys,
            Yvars,
            task_features,
            fidelity_features,
            metric_names,
            state_dict,
            **kwargs,
        ) -> model

    Here `Xs`, `Ys`, `Yvars` are lists of tensors (one element per outcome),
    `task_features` identifies columns of Xs that should be modeled as a task,
    `fidelity_features` is a list of ints that specify the positions of fidelity
    parameters in 'Xs', `metric_names` provides the names of each `Y` in `Ys`,
    `state_dict` is a pytorch module state dict, and `model` is a BoTorch `Model`.
    Optional kwargs are being passed through from the `LegacyBoTorchGenerator`
    constructor. This callable is assumed to return a fitted BoTorch model that has
    the same dtype and lives on the same device as the input tensors.

    ::

        model_predictor(model, X) -> [mean, cov]

    Here `model` is a fitted botorch model, `X` is a tensor of candidate points,
    and `mean` and `cov` are the posterior mean and covariance, respectively.

    ::

        acqf_constructor(
            model,
            objective_weights,
            outcome_constraints,
            X_observed,
            X_pending,
            **kwargs,
        ) -> acq_function


    Here `model` is a botorch `Model`, `objective_weights` is a tensor of weights
    for the model outputs, `outcome_constraints` is a tuple of tensors describing
    the (linear) outcome constraints, `X_observed` are previously observed points,
    and `X_pending` are points whose evaluation is pending. `acq_function` is a
    BoTorch acquisition function crafted from these inputs. For additional
    details on the arguments, see `get_qLogNEI`.

    ::

        acqf_optimizer(
            acq_function,
            bounds,
            n,
            inequality_constraints,
            equality_constraints,
            fixed_features,
            rounding_func,
            **kwargs,
        ) -> candidates

    Here `acq_function` is a BoTorch `AcquisitionFunction`, `bounds` is a tensor
    containing bounds on the parameters, `n` is the number of candidates to be
    generated, `inequality_constraints` are inequality constraints on parameter
    values, `fixed_features` specifies features that should be fixed during
    generation, and `rounding_func` is a callback that rounds an optimization
    result appropriately. `candidates` is a tensor of generated candidates.
    For additional details on the arguments, see `scipy_optimizer`.

    ::

        best_point_recommender(
            model,
            bounds,
            objective_weights,
            outcome_constraints,
            linear_constraints,
            fixed_features,
            model_gen_options,
            target_fidelities,
        ) -> candidates

    Here `model` is a TorchGenerator, `bounds` is a list of tuples containing bounds
    on the parameters, `objective_weights` is a tensor of weights for the model outputs,
    `outcome_constraints` is a tuple of tensors describing the (linear) outcome
    constraints, `linear_constraints` is a tuple of tensors describing constraints
    on the design, `fixed_features` specifies features that should be fixed during
    generation, `model_gen_options` is a config dictionary that can contain
    model-specific options, and `target_fidelities` is a map from fidelity feature
    column indices to their respective target fidelities, used for multi-fidelity
    optimization problems. % TODO: refer to an example.
    """

    dtype: torch.dtype | None
    device: torch.device | None
    Xs: list[Tensor]
    Ys: list[Tensor]
    Yvars: list[Tensor]
    _model: Model | None
    _search_space_digest: SearchSpaceDigest | None = None

    def __init__(
        self,
        model_constructor: TModelConstructor = get_and_fit_model,
        model_predictor: TModelPredictor = predict_from_model,
        acqf_constructor: TAcqfConstructor = get_qLogNEI,
        # pyre-fixme[9]: acqf_optimizer declared/used type mismatch
        acqf_optimizer: TOptimizer = scipy_optimizer,
        best_point_recommender: TBestPointRecommender = recommend_best_observed_point,
        refit_on_cv: bool = False,
        warm_start_refitting: bool = True,
        use_input_warping: bool = False,
        use_loocv_pseudo_likelihood: bool = False,
        prior: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        warnings.warn(
            "The legacy `LegacyBoTorchGenerator` and its subclasses, including the "
            f"current class `{self.__class__.__name__}`, slated for deprecation. "
            "These models will not be supported going forward and may be "
            "fully removed in a future release. Please consider using the "
            "Modular BoTorch Generator (MBG) setup (ax/models/torch/botorch_modular) "
            "instead. If you run into a use case that is not supported by MBG, "
            "please raise this with an issue at https://github.com/facebook/Ax",
            DeprecationWarning,
            stacklevel=2,
        )
        self.model_constructor = model_constructor
        self.model_predictor = model_predictor
        self.acqf_constructor = acqf_constructor
        self.acqf_optimizer = acqf_optimizer
        self.best_point_recommender = best_point_recommender
        # pyre-fixme[4]: Attribute must be annotated.
        self._kwargs = kwargs
        self.refit_on_cv = refit_on_cv
        self.warm_start_refitting = warm_start_refitting
        self.use_input_warping = use_input_warping
        self.use_loocv_pseudo_likelihood = use_loocv_pseudo_likelihood
        self.prior = prior
        self._model: Model | None = None
        self.Xs = []
        self.Ys = []
        self.Yvars = []
        self.dtype = None
        self.device = None
        self.task_features: list[int] = []
        self.fidelity_features: list[int] = []
        self.metric_names: list[str] = []

    @copy_doc(TorchGenerator.fit)
    def fit(
        self,
        datasets: Sequence[SupervisedDataset],
        search_space_digest: SearchSpaceDigest,
        candidate_metadata: list[list[TCandidateMetadata]] | None = None,
    ) -> None:
        if len(datasets) == 0:
            raise DataRequiredError(
                "LegacyBoTorchGenerator.fit requires non-empty data sets."
            )
        self.Xs, self.Ys, self.Yvars = _datasets_to_legacy_inputs(datasets=datasets)
        self.metric_names = sum((ds.outcome_names for ds in datasets), [])
        # Store search space info for later use (e.g. during generation)
        self._search_space_digest = search_space_digest
        self.dtype = self.Xs[0].dtype
        self.device = self.Xs[0].device
        self.task_features = none_throws(
            normalize_indices(search_space_digest.task_features, d=self.Xs[0].size(-1))
        )
        self.fidelity_features = none_throws(
            normalize_indices(
                search_space_digest.fidelity_features, d=self.Xs[0].size(-1)
            )
        )
        extra_kwargs = {} if self.prior is None else {"prior": self.prior}
        self._model = self.model_constructor(  # pyre-ignore [28]
            Xs=self.Xs,
            Ys=self.Ys,
            Yvars=self.Yvars,
            task_features=self.task_features,
            fidelity_features=self.fidelity_features,
            metric_names=self.metric_names,
            use_input_warping=self.use_input_warping,
            use_loocv_pseudo_likelihood=self.use_loocv_pseudo_likelihood,
            **extra_kwargs,
            **self._kwargs,
        )

    @copy_doc(TorchGenerator.predict)
    def predict(
        self, X: Tensor, use_posterior_predictive: bool = False
    ) -> tuple[Tensor, Tensor]:
        if use_posterior_predictive:
            warnings.warn(
                f"{self.__class__.__name__} does not support posterior-predictive "
                "predictions. Ignoring `use_posterior_predictive`. ",
                AxWarning,
                stacklevel=2,
            )
        return self.model_predictor(model=self.model, X=X)  # pyre-ignore [28]

    @copy_doc(TorchGenerator.gen)
    def gen(
        self,
        n: int,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
    ) -> TorchGenResults:
        options = torch_opt_config.model_gen_options or {}
        acf_options = options.get(Keys.ACQF_KWARGS, {})
        optimizer_options = options.get(Keys.OPTIMIZER_KWARGS, {})

        if search_space_digest.fidelity_features:
            raise NotImplementedError(
                "Base LegacyBoTorchGenerator does not support fidelity_features."
            )
        X_pending, X_observed = _get_X_pending_and_observed(
            Xs=self.Xs,
            objective_weights=torch_opt_config.objective_weights,
            bounds=search_space_digest.bounds,
            pending_observations=torch_opt_config.pending_observations,
            outcome_constraints=torch_opt_config.outcome_constraints,
            linear_constraints=torch_opt_config.linear_constraints,
            fixed_features=torch_opt_config.fixed_features,
            fit_out_of_design=torch_opt_config.fit_out_of_design,
        )
        model = self.model
        # subset model only to the outcomes we need for the optimization	357
        if options.get(Keys.SUBSET_MODEL, True):
            subset_model_results = subset_model(
                model=model,
                objective_weights=torch_opt_config.objective_weights,
                outcome_constraints=torch_opt_config.outcome_constraints,
            )
            model = subset_model_results.model
            objective_weights = subset_model_results.objective_weights
            outcome_constraints = subset_model_results.outcome_constraints
        else:
            objective_weights = torch_opt_config.objective_weights
            outcome_constraints = torch_opt_config.outcome_constraints

        bounds_ = torch.tensor(
            search_space_digest.bounds, dtype=self.dtype, device=self.device
        )
        bounds_ = bounds_.transpose(0, 1)

        botorch_rounding_func = get_rounding_func(torch_opt_config.rounding_func)

        from botorch.exceptions.errors import UnsupportedError

        # pyre-fixme[53]: Captured variable `X_observed` is not annotated.
        # pyre-fixme[53]: Captured variable `X_pending` is not annotated.
        # pyre-fixme[53]: Captured variable `acf_options` is not annotated.
        # pyre-fixme[53]: Captured variable `botorch_rounding_func` is not annotated.
        # pyre-fixme[53]: Captured variable `bounds_` is not annotated.
        # pyre-fixme[53]: Captured variable `model` is not annotated.
        # pyre-fixme[53]: Captured variable `objective_weights` is not annotated.
        # pyre-fixme[53]: Captured variable `optimizer_options` is not annotated.
        # pyre-fixme[53]: Captured variable `outcome_constraints` is not annotated.
        def make_and_optimize_acqf(override_qmc: bool = False) -> tuple[Tensor, Tensor]:
            add_kwargs = {"qmc": False} if override_qmc else {}
            acquisition_function = self.acqf_constructor(
                model=model,
                objective_weights=objective_weights,
                outcome_constraints=outcome_constraints,
                X_observed=X_observed,
                X_pending=X_pending,
                **acf_options,
                **add_kwargs,
            )
            acquisition_function = assert_is_instance(
                acquisition_function, AcquisitionFunction
            )
            # pyre-ignore: [28]
            candidates, expected_acquisition_value = self.acqf_optimizer(
                acq_function=assert_is_instance(
                    acquisition_function, AcquisitionFunction
                ),
                bounds=bounds_,
                n=n,
                inequality_constraints=_to_inequality_constraints(
                    linear_constraints=torch_opt_config.linear_constraints
                ),
                fixed_features=torch_opt_config.fixed_features,
                rounding_func=botorch_rounding_func,
                **optimizer_options,
            )
            return candidates, expected_acquisition_value

        try:
            candidates, expected_acquisition_value = make_and_optimize_acqf()
        except UnsupportedError as e:  # untested
            if "SobolQMCSampler only supports dimensions" in str(e):
                # dimension too large for Sobol, let's use IID
                candidates, expected_acquisition_value = make_and_optimize_acqf(
                    override_qmc=True
                )
            else:
                raise e

        gen_metadata = {}
        if expected_acquisition_value.numel() > 0:
            gen_metadata["expected_acquisition_value"] = (
                expected_acquisition_value.tolist()
            )

        return TorchGenResults(
            points=candidates.detach().cpu(),
            weights=torch.ones(n, dtype=self.dtype),
            gen_metadata=gen_metadata,
        )

    @copy_doc(TorchGenerator.best_point)
    def best_point(
        self,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
    ) -> Tensor | None:
        if torch_opt_config.is_moo:
            raise NotImplementedError(
                "Best observed point is incompatible with MOO problems."
            )
        target_fidelities = {
            k: v
            for k, v in search_space_digest.target_values.items()
            if k in search_space_digest.fidelity_features
        }
        return self.best_point_recommender(  # pyre-ignore [28]
            model=self,
            bounds=search_space_digest.bounds,
            objective_weights=torch_opt_config.objective_weights,
            outcome_constraints=torch_opt_config.outcome_constraints,
            linear_constraints=torch_opt_config.linear_constraints,
            fixed_features=torch_opt_config.fixed_features,
            model_gen_options=torch_opt_config.model_gen_options,
            target_fidelities=target_fidelities,
        )

    @copy_doc(TorchGenerator.cross_validate)
    def cross_validate(
        self,
        datasets: Sequence[SupervisedDataset],
        X_test: Tensor,
        search_space_digest: SearchSpaceDigest,
        use_posterior_predictive: bool = False,
    ) -> tuple[Tensor, Tensor]:
        if self._model is None:
            raise RuntimeError("Cannot cross-validate model that has not been fitted.")
        if self.refit_on_cv:
            state_dict = None
        else:
            state_dict = deepcopy(self.model.state_dict())
        Xs, Ys, Yvars = _datasets_to_legacy_inputs(datasets=datasets)
        model = self.model_constructor(  # pyre-ignore: [28]
            Xs=Xs,
            Ys=Ys,
            Yvars=Yvars,
            task_features=self.task_features,
            state_dict=state_dict,
            fidelity_features=self.fidelity_features,
            metric_names=self.metric_names,
            refit_model=self.refit_on_cv,
            use_input_warping=self.use_input_warping,
            use_loocv_pseudo_likelihood=self.use_loocv_pseudo_likelihood,
            **self._kwargs,
        )
        # pyre-ignore: [28]
        return self.model_predictor(
            model=model, X=X_test, use_posterior_predictive=use_posterior_predictive
        )

    def feature_importances(self) -> npt.NDArray:
        return get_feature_importances_from_botorch_model(model=self._model)

    @property
    def search_space_digest(self) -> SearchSpaceDigest:
        if self._search_space_digest is None:
            raise RuntimeError(
                "`search_space_digest` is not initialized. Please fit the model first."
            )
        return self._search_space_digest

    @search_space_digest.setter
    def search_space_digest(self, value: SearchSpaceDigest) -> None:
        raise RuntimeError("Setting search_space_digest manually is disallowed.")

    @property
    def model(self) -> Model:
        if self._model is None:
            raise RuntimeError(
                "`model` is not initialized. Please fit the model first."
            )
        return self._model

    @model.setter
    def model(self, model: Model) -> None:
        self._model = model  # there are a few places that set model directly


def get_rounding_func(
    rounding_func: Callable[[Tensor], Tensor] | None,
) -> Callable[[Tensor], Tensor] | None:
    if rounding_func is None:
        botorch_rounding_func = rounding_func
    else:
        # make sure rounding_func is properly applied to q- and t-batches
        def botorch_rounding_func(X: Tensor) -> Tensor:
            batch_shape, d = X.shape[:-1], X.shape[-1]
            X_round = torch.stack(
                [rounding_func(x) for x in X.view(-1, d)]  # pyre-ignore: [16]
            )
            return X_round.view(*batch_shape, d)

    return botorch_rounding_func


def get_feature_importances_from_botorch_model(
    model: Model | ModuleList | None,
) -> npt.NDArray:
    """Get feature importances from a list of BoTorch models.

    Args:
        models: BoTorch model to get feature importances from.

    Returns:
        The feature importances as a numpy array where each row sums to 1.
    """
    if model is None:
        raise RuntimeError(
            "Cannot calculate feature_importances without a fitted model."
            "Call `fit` first."
        )
    elif isinstance(model, ModelList):
        models = model.models
    else:
        models = [model]
    lengthscales = []
    for m in models:
        try:
            # this can be a ModelList of a SAAS and STGP, so this is a necessary way
            # to get the lengthscale
            if hasattr(m.covar_module, "base_kernel"):
                # pyre-fixme[16]: Undefined attribute: Item `torch._tensor.Tensor` of...
                ls = m.covar_module.base_kernel.lengthscale
            else:
                # pyre-fixme[16]: Undefined attribute: Item `torch._tensor.Tensor` of...
                ls = m.covar_module.lengthscale
        except AttributeError:
            ls = None
        # pyre-fixme[29]: Call error: `typing.Union[BoundMethod[typing.Callable(torch...
        if ls is None or ls.shape[-1] != m.train_inputs[0].shape[-1]:
            # TODO: We could potentially set the feature importances to NaN in this
            # case, but this require knowing the batch dimension of this model.
            # Consider supporting in the future.
            raise NotImplementedError(
                "Failed to extract lengthscales from `m.covar_module` "
                "and `m.covar_module.base_kernel`"
            )
        if ls.ndim == 2:
            ls = ls.unsqueeze(0)
        # pyre-fixme[6]: Incompatible parameter type: In call `is_ensemble`, for 1st ...
        if is_ensemble(m):  # Take the median over the model batch dimension
            ls = torch.quantile(ls, q=0.5, dim=0, keepdim=True)
        lengthscales.append(ls)
    lengthscales = torch.cat(lengthscales, dim=0)
    feature_importances = (1 / lengthscales).detach().cpu()  # pyre-ignore
    # Make sure the sum of feature importances is 1.0 for each metric
    feature_importances /= feature_importances.sum(dim=-1, keepdim=True)
    return feature_importances.numpy()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import warnings

import numpy as np
import numpy.typing as npt
from ax.exceptions.core import UserInputError
from scipy.stats import norm


class Ashr:
    r"""
    An empirical Bayes model for estimating the effect sizes. Given the observations
    Y_i and their variances Yvar_i, Ashr estimates the underlying effect sizes mu_i
    by placing a mixture of Gaussians prior on the effect sizes. The prior consists of
    a point mass at zero and a set of centered Gaussians with specified variaces.
    The mixture proportions in the prior as well as the variances of the Gaussians
    in the mixture are learned based on observed outcome data.
    Methodology is based on the paper: False discovery rates: a new deal by
    M. Stephens https://academic.oup.com/biostatistics/article/18/2/275/2557030.
    """

    def __init__(
        self,
        Y: npt.NDArray,
        Yvar: npt.NDArray,
        prior_vars: npt.NDArray | None = None,
        eb_grid_param: float = 2.0,
    ) -> None:
        r"""
        Args:
            Y: A length n array denoting the observed treatment effects.
            Yvar: A length n array denoting the variances
                of the observed values.
            prior_vars: A length k array denoting the variances
                of normal distributions in the mixture of Gaussians prior.
                In case of None, the variances are estimated based on data
                using the provided grid parameter.
            eb_grid_param: A grid parameter for estimating the prior variances
                based on data in case none were given.
        """
        self.Y: npt.NDArray = Y
        self.Yvar: npt.NDArray = Yvar
        if prior_vars is None:
            prior_stds = prior_grid(Y=Y, Yvar=Yvar, grid_param=eb_grid_param)
            prior_vars = prior_stds**2
        self.prior_vars: npt.NDArray = prior_vars
        self.ll: npt.NDArray = marginal_densities(Y=Y, Yvar=Yvar, prior_vars=prior_vars)

    def posterior(self, w: npt.NDArray) -> GaussianMixture:
        r"""
        The posterior for mu_i can be calculated via the following rules.
        For the normal prior mu_i~N(0, sigma_k^2) and
        likelihood hat{mu}_i~N(mu_i, s_i^2),
        posterior for mu_i is N(sigma_1^2*hat{mu}_i/s_i*2, sigma_1^2),
        where sigma_1^2=sigma_k^2*s_i^2/(sigma_k^2+s_i^2).

        Args:
            w: (n,k) dim matrix containing the weights of the posterior mixture.

        Returns:
            Mixutre of Gaussians distribution.
        """

        # (n, k) matrix of normal variances
        # a variance per mixture component and per observation
        normal_vars_by_class = np.divide(
            np.multiply.outer(self.Yvar, self.prior_vars),
            np.add.outer(self.Yvar, self.prior_vars),
        )
        # (n, k) matrix of normal standard deviations
        # a mean per mixture component and per observation
        normal_means_by_class = normal_vars_by_class * (self.Y / self.Yvar)[:, None]

        return GaussianMixture(
            normal_means_by_class=normal_means_by_class,
            normal_vars_by_class=normal_vars_by_class,
            weights=w,
        )

    def fit(
        self,
        lambdas: npt.NDArray | None = None,
        threshold: float = 10e-4,
        nsteps: int = 1000,
    ) -> dict[str, npt.NDArray]:
        r"""
        Fit Ashr to estimate prior proportions pi, posterior weights and lfdr.

        Args:
            lambdas: A length k array of penalty levels corresponding to each of
                k classes with entry equal to one meaning no penalty
                for the corresponding class.
            thereshold: The threshold used in the EM stoping rule.
                If the difference between two consecutive estimates of
                weights in the EM algorithm is smaller than the threshold,
                the algorithm stops.
            nsteps: The maximum number of steps in the EM algorithm.

        Returns:
            A dict containing
            - weights: (n, k)-dim estimated weight matrix for computing posteriors,
            - pi: parameter estimates: (n, k) dim matrix of proportions, and
            - lfdr: length n local False Discovery Rate (FDR) sequence. lfdr equals
                the posterior probability of true parameter being zero.
        """
        k = len(self.prior_vars)  # total number of classes
        if lambdas is None:
            lambdas = np.ones(k)  # no penalty
        if len(lambdas) != k:
            raise ValueError(
                "The length of the penalty sequence should be the number of "
                "prior classes."
            )
        lambdas = lambdas.astype(np.float64)

        results = fit_ashr_em(
            ll=self.ll, lambdas=lambdas, threshold=threshold, nsteps=nsteps
        )

        return results


class GaussianMixture:
    r"""
    A weighted mixure of Gaussian distributions. The class computes the mean,
    standard errors and tails of each of n random variables
    from a mixture of k Gaussians.
    The Gaussians in the mixture are allowed to be degenerate,
    i.e., have zero variance. This is used in the Ashr model since one of the
    distributions in the prior mixture is a point mass at zero.
    """

    def __init__(
        self,
        normal_means_by_class: npt.NDArray,
        normal_vars_by_class: npt.NDArray,
        weights: npt.NDArray,
    ) -> None:
        r"""
        Args:
            normal_means_by_class: (n, k)-dim matrix of normal means for
                each of the classes. Each of the n random variables comes from
                a mixture of k normal distributions.
            normal_var_by_class: (n, k)-dim matrix of normal variances
                for each of the k classes. The first column is all zeros,
                the variance of the null class.
            weights: (n, k)-dim matrix of weights on individual distributions
                per prior class.
        """
        self.normal_means_by_class = normal_means_by_class
        self.normal_vars_by_class = normal_vars_by_class
        self.weights = weights

    def tail_probabilities(
        self, left_tail: bool = True, threshold: float = 0.0
    ) -> npt.NDArray:
        r"""
        Args:
            left_tail: An indicator for the tail probability to calculate.
                Note that neither tail includes null class.

            threshold: For left tail, the returned value measures probability of the
                effect being less than the threshold. For right tail, it is the
                probability of the effect being larger than the threshold.

        Returns:
            Length n array of tail probabilities for each of n rvs.
        """
        tails_by_class = np.zeros_like(self.normal_means_by_class)

        # normal left tails
        for i in range(tails_by_class.shape[0]):
            for j in range(tails_by_class.shape[1]):
                tails_by_class[i, j] = (
                    norm.cdf(
                        -np.divide(
                            self.normal_means_by_class[i, j] - threshold,
                            np.sqrt(self.normal_vars_by_class[i, j]),
                        )
                    )
                    if self.normal_vars_by_class[i, j] > 0
                    else 0.0
                )
                # correcting the normal tails in case of a right tail
                if left_tail is False and self.normal_vars_by_class[i, j] > 0:
                    tails_by_class[i, j] = 1.0 - tails_by_class[i, j]

        return np.multiply(tails_by_class, self.weights).sum(axis=1)

    @property
    def means(self) -> npt.NDArray:
        r"""
        Returns:
            Length n array of final means for each rv.
        """
        return np.multiply(self.weights, self.normal_means_by_class).sum(axis=1)

    @property
    def vars(self) -> npt.NDArray:
        r""" "
        Returns:
            Length n array of final standard deviations for each effect.
        """
        # standard errors of the mixture distributions
        # https://en.wikipedia.org/wiki/Mixture_distribution#Moments
        return (
            np.multiply(
                self.weights,
                self.normal_means_by_class**2 + self.normal_vars_by_class,
            ).sum(axis=1)
            - self.means**2
        )


def prior_grid(
    Y: npt.NDArray, Yvar: npt.NDArray, grid_param: float = 2.0
) -> npt.NDArray:
    r"""
    Produces the grid of standard deviations for each of the Gaussians in the prior
    mixture based on the observed data.

    Args:
        Y: A length n array of the observed treatment effects.
        Yvar: A length n array of observed variances of the above observations.
        grid_param: A grid parameter. Default 2.0 recommended in the paper to control
            the number of Gaussians in the mixture.

    Returns:
        A length n array of the standard deviations of the centered Gaussians in the
        mixture of Gaussians prior.
    """
    m = np.sqrt(grid_param)
    sigma_lower = np.min(np.sqrt(Yvar)) / 10.0
    sigma_upper = np.max(Y**2 - Yvar)
    sigma_upper = 2 * np.sqrt(sigma_upper) if sigma_upper > 0 else 8 * sigma_lower
    max_power = int(np.ceil(np.log(sigma_upper / sigma_lower) / np.log(m)))
    return np.array([0] + [sigma_lower * (m**power) for power in range(max_power + 1)])


def marginal_densities(
    Y: npt.NDArray,
    Yvar: npt.NDArray,
    prior_vars: npt.NDArray,
) -> npt.NDArray:
    r"""
    Evaluates marginal densities for each observed statistics and each prior class.

    Args:
        Y: A length n array denoting the observed treatment effects.
        Yvar: A length n array denoting the standard variances
            of the observed values.
        prior_vars: A length k array denoting the variances
            of prior classes.

    Returns:
        (n, k) dim matrix ll, where ll_{jk} is marginal density of j-th statistics
        eval at its observed value, assuming prior is coming from k-th class.
    """
    k = len(prior_vars)  # total number of classes
    n = len(Y)  # total number of observations
    if prior_vars[0] != 0:
        raise UserInputError(
            "Ashr prior consists of a mixture of Gaussians where the "
            "first Gaussian in the prior mixture should be a point mass at zero. \
            This degenerate Gaussian represents the prior on the effects being null."
        )
    ll = np.zeros((n, k), dtype=np.float64)

    # marginal densities when prior is mass at zero
    ll[:, 0] = norm.pdf(Y, loc=0, scale=np.sqrt(Yvar))

    for i in range(1, k):
        ll[:, i] = norm.pdf(
            Y,
            loc=0.0,
            scale=np.sqrt(prior_vars[i] + Yvar),
        )
    return ll


def compute_weights(ll: npt.NDArray, pi: npt.NDArray) -> npt.NDArray:
    r"""
    Compute posterior weights based on marginal densities and prior probabilities.

    Args:
        ll: (n,k)-dim matrix of marginal densities eval at each observation,
        pi: length k vector of prior mixture proportions.

    Returns:
        (n,k)-dim matrix of weights.
    """
    # multiply each row of ll with the corresponding element of pi vector
    w = np.multiply(ll, pi)
    # divide weights by the sum across each row
    w = w / w.sum(axis=1, keepdims=True)
    return w


def fit_ashr_em(
    ll: npt.NDArray,
    lambdas: npt.NDArray,
    threshold: float = 10e-4,
    nsteps: int = 1000,
) -> dict[str, npt.NDArray]:
    r"""
    Estimating proportions and posterior weights via an
    Expectation Maximization (EM) algorithm.

    Args:
        ll: (n,k)-dim matrix of marginal densities eval at each observation,
            marginalizing over the true effects.
        lambdas: A length k array of penalty levels.
        thereshold: If the difference between two consecutive estimates of
            weights is smaller than the threshold, the algorithm stops.
        nsteps: The maximum number of steps in the EM algorithm.

    Returns:
        A dictionary containing:
        - weights: (n,k)-dim matrix of weights,
        - pi: length k vector of estimates of prior mixture proportions, and
       - lfdr: length n local False Discovery Rate (FDR) sequence. lfdr equals
            the posterior probability of true parameter being zero.
    """
    n, k = ll.shape
    # initializing pi vector
    if k - 1 < n:
        pi = np.ones(k) / n
        pi[0] = 1 - (k - 1) / n
    else:
        pi = np.ones(k) / k

    w = np.zeros_like(ll)

    for _ in range(nsteps):
        # E-step: compute weight matrix w; size of w: (n, k)
        w = compute_weights(ll=ll, pi=pi)
        # M-step: update pi
        ns = w.sum(axis=0).squeeze() + lambdas - 1.0  # length k
        pi_new = ns / ns.sum()  # length k
        if sum(abs(pi - pi_new)) <= threshold:
            w = compute_weights(ll=ll, pi=pi_new)
            return {"weights": w, "pi": pi_new, "lfdr": w[:, 0]}
        pi = pi_new

    warnings.warn("EM did not converge.", stacklevel=2)
    return {"weights": w, "pi": pi, "lfdr": w[:, 0]}
